;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "entities.h.s"
.include "../system/ai.h.s"
.include "../sprites/sprites.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables                                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
entities::          .ds entity_size*entity_max_count
entity_array_end:   .db 0x00
entity_array_free_space:        .db entity_max_count
entity_array_next_free_position: .dw #entities

.globl cpct_memset_asm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Entity                                                    ;;
;;  Creates an entity using the specified template                  ;;
;; Input: HL with the template to use (memory position)             ;;
;; Output:  IX with new entity's mem pos, A = 0 if no new entity    ;;
;;          Destroys BC and DE                                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_create_one_entity:
    ;; Check if there is free space left
    ld a, (entity_array_free_space)
    or a
    jr z, man_create_one_skip

    ;; Search next free position
    ld ix, (entity_array_next_free_position)

    ;; Load template on new entity
    ;; HL points to template array
    push ix                     ;; Use the pile to pass data to DE
    pop de                      ;; DE points to the start of the new entity
    ld bc, #entity_size         ;; BC holds bytes to copy
    ;; HL                       ;; HL holds the start of the template
    ldir

    ;; Change next free position to [entity size] onwards
    ;; Also leave IX pointing to new entity
    push ix             ;; Save new entity's position
    ld bc, #entity_size ;; Load the offset to reach next entity's start
    add ix, bc          ;; Add it to IX
    ld (entity_array_next_free_position), ix ;; Save it
    pop ix              ;; Recover new entity's position

    ;; Decrease free space counter
    ld a, (entity_array_free_space)
    dec a
    ld (entity_array_free_space), a

    
    ld a, #0x01 ;; Load A output (= 1 if new entity was created)
    jr man_create_one_end

    ;; Label for creation skipping if there is no free space left
    man_create_one_skip:
    ld a, #0x00 ;; Load A output (= 0 if NO new entity was created)

    man_create_one_end:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Destroy Entity                                                   ;;
;;  Destroys the specified entity, and reorganizes the array        ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_destroy_one_entity:
    ;; Execute death action of this entity
    ld l, ent_death_beh(ix)
    ld h, 1+ent_death_beh(ix)
    ld de, #destroy_one_return_point
    push de     ;; Manual CALL
    jp (hl)     ;; Jump to death routine
    destroy_one_return_point:

    ;; Load HL with last entity's memory position
    ld hl, (entity_array_next_free_position);; HL stores the origin
    ld a, #entity_size
    destroy_one_go_back_an_entity:
        dec hl
        dec a
    jr nz, destroy_one_go_back_an_entity

    ;; Save next free position
    ld (entity_array_next_free_position), hl

    ;; Copy last entity into dead entity
    push ix
    pop de                                  ;; DE must store the destiny (Dead entity)
    ld bc, #entity_size                     ;; BC holds the number of bytes to copy
    ;; HL                                   ;; HL holds the origin of the entity to copy
    ldir

    ;; Mark the moved entity's old position as free
    ld hl, (entity_array_next_free_position) 
    ld (hl), #0x00 ;; Mark as invalid  

    ;; Update entity's array free space
    ld hl, #entity_array_free_space
    inc (hl)
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mark Entity for Destruction                                      ;;
;;  Marks the specified entity for later destruction                ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A                                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_mark_one_entity_for_destruction:
    ;; Mark an entity as dead using the dead flag
    ld a, ent_type(ix)  ;; Load the type
    or #ent_type_dead   ;; Add death mask bit to A
    ld ent_type(ix), a  ;; Save that on type
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mark Entity for Destruction of Type                              ;;
;;  Marks all hud items of the specified type for later destruction ;;
;; Input: B with filter mask                                        ;;
;; Output: Destroys IX, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_mark_for_destroy_entity_of_type:
    ;; Load entity array start
    ld ix, #entities

    ;; Loop until it finds an invalid entity
    entities_mark_type_loop:
        ;; Check if entity is valid
        ld a, ent_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        ret z               ;; If its zero (invalid) finsih te loop

        ;; Check if is one of the types we want to destroy
        ;;ld a, ent_type(ix)    ;; A = type
        and b                   ;; B = filter mask
        jr z, ent_mark_type_destruction_skip   ;; If zero, no filter passed, skip death

        ;; If its valid, destroy it
        call man_mark_one_entity_for_destruction
        ent_mark_type_destruction_skip:

        ;; Move to next hud item
        ld de, #entity_size
        add ix, de

    ;; Loop end
    jr entities_mark_type_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Dead Entites                                              ;;
;;  Search the array for marked-as-dead entites to destroy them     ;;
;;  at once                                                         ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B and DE                                 ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_update_dead_entities:
    ;; Load entity array start
    ld ix, #entities

    ;; Loop until it finds an invalid entity
    man_update_entities_loop:
        ;; Check if entity is valid
        ld a, ent_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        ret z               ;; If its zero (invalid) finsih te loop

        ;; Check if entity is marked for death
        ;; A still stores the type
        ld b, #ent_type_dead;; Load dead mask
        and b               ;; Bitwise AND with death mask is Zero if death flag is Zero (not dead)
        jr z, man_update_entities_death_skip

        ;; Kill the entity (IX is the input)
        call man_destroy_one_entity
        jr man_update_entities_increase_skip    ;; If we destroyed one, we do not advance to the next position
        
        ;; Move to next entity
        man_update_entities_death_skip: ;; Label for death skip
        ld de, #entity_size
        add ix, de

        man_update_entities_increase_skip:

    ;; Loop end
    jr man_update_entities_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply to All Matching                                            ;;
;;  Apply the specified function to every entity that matches the   ;;
;;  filter mask                                                     ;;
;; Input: HL with routine's memory pos, B with the filter mask      ;;
;; Output: Destroys IX, A and DE. Might destroy F                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_for_all_matching:
    ;; Load entity array start
    ld ix, #entities

    ;; Loop until it finds an invalid entity
    man_for_all_matching_loop:
        ;; Check if entity is valid
        ld a, ent_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        ret z               ;; If its zero (invalid) finsih te loop

        ;; Check if entity has the desired components
        ld a, ent_comps(ix) ;; Load component mask
        and b               ;; Bitwise AND with filter mask
        sub b               ;; Result (in A) must be equal to B, so A - B = 0
        jr nz, man_for_all_matching_routine_skip

        ;; Save input parameters data
        push hl
        push bc

        ;; If its valid, apply routine
        ld de, #man_for_all_matching_return_point   ;; Load return position
        push de                                     ;; Push that position to the pile
        jp (hl)                                     ;; CALL simulated with manual push and jump

        ;; Load input parameters data
        man_for_all_matching_return_point:
        pop bc
        pop hl

        man_for_all_matching_routine_skip:
        ;; Move to next entity
        ld de, #entity_size
        add ix, de

    ;; Loop end
    jr man_for_all_matching_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply to All Pairs Matching                                      ;;
;;  Apply the specified function to every pairs of entities that    ;;
;;  matches the filter mask                                         ;;
;; Input: HL with routine's memory pos, B with the filter mask      ;;
;; Output: Destroys IX, IY, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_for_all_pairs_matching:
    ;; Load entity array start
    ld ix, #entities

    ;; Loop until it finds an invalid entity
    pairs_outer_loop:
        ;; Check if left entity is valid
        ld a, ent_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        ret z               ;; If its zero (invalid) finsih te loop

        ;; Check if left entity has the desired components
        ld a, ent_comps(ix) ;; Load component mask
        and b               ;; Bitwise AND with filter mask
        sub b               ;; Result (in A) must be equal to B, so A - B = 0
        jr nz, pairs_end_inner_loop

            ;;; RIGHT ENTITY
            ;;;;;;;;;;;;;;;;

            ;; Make pointer to right entity
            push ix             ;; Copy left entity ptr to right ptr
            pop iy
            ld de, #entity_size ;; Prepare to add [entity_size] to right ptr
            add iy, de          ;; Move to next entity

            ;; Label for loop repeating
            pairs_inner_loop:

            ;; Check if right entity is valid
            ld a, ent_type(iy)  ;; Load the type in A
            or a                ;; Check zero value
            jr z, pairs_end_inner_loop ;; If its zero (invalid) finsih te loop

            ;; Check if right entity has the desired components
            ld a, ent_comps(iy) ;; Load component mask
            and b               ;; Bitwise AND with filter mask
            sub b               ;; Result (in A) must be equal to B, so A - B = 0
            jr nz, pairs_routine_skip

                ;; Save input parameters data
                push hl
                push bc

                ;; If its valid, apply routine
                ld de, #pairs_return_point ;; Load return position
                push de                                         ;; Push that position to the pile
                jp (hl)                                         ;; CALL simulated with manual push and jump
                
                ;; Load input parameters data
                pairs_return_point:
                pop bc
                pop hl

                ;; Label for skipping routine apply
                pairs_routine_skip:

            ;; Move right ptr to next entity
            ld de, #entity_size
            add iy, de

            jr pairs_inner_loop

        pairs_end_inner_loop:
        ;; Move to next entity
        ld de, #entity_size
        add ix, de

    ;; Loop end
    jr pairs_outer_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Search a Matching Entity                                         ;;
;;  Search for an specific entity type                              ;;
;; Input: B with the filter mask                                    ;;
;; Output: Z if A=0 (not found) or NZ if A=1 (found)                ;;
;;         IX with first coincience                                ;;
;;         Destroys DE                                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_search_type:
    ;; Load entity array start
    ld ix, #entities

    ;; Loop until it finds an invalid entity
    man_search_type_loop:
        ;; Check if entity is valid
        ld a, ent_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        jr z, man_search_one_not_found ;; If its zero (invalid) finsih te loop

        ;; Check if entity has the desired components
        ;; A = type mask    ;; A stores type mask
        and b               ;; Bitwise AND with type filter mask
        jr nz, man_search_one_found_it

        ;; Move to next entity
        ld de, #entity_size
        add ix, de

    ;; Loop end
    jr man_search_type_loop

    ;; Label for succesful search
    man_search_one_found_it:
    ld a, #0x01
    jr man_search_one_end

    ;; Label for search fail, array loop finished
    man_search_one_not_found:
    ld a, #0x00

    man_search_one_end:
    cp #0 ;; Z if A=0 (not found) or NZ if A=1 (found)
ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Empty the entities array                                         ;;
;;  Sets the entities array to 00                                   ;;
;; Input: -                                                         ;;
;; Output: Destroys A, BC and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_empty_entities:
    ;; load in DE the start of the entities array
    ld de, #entities
    ;; load in A 0 to fill the array
    ld a, #0x00
    ;; load in BC the size in BYTES
    ld bc, #entities_array_byte_size
    ;; do the same as in DE
    ld hl, #entities
    ;;the next free entity is the begining of the array yay
    ld (entity_array_next_free_position), hl
    ;;fill array with 0s
    call cpct_memset_asm
    ;;load in A the max entities cap
    ld a, #entity_max_count
    ;;set the free space to the max count
    ld (entity_array_free_space), a
ret
