;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "game.h.s"
.include "cpctelera.h.s"
.include "entities.h.s"
.include "huditems.h.s"
.include "level_reader.h.s"
.include "bullet.h.s"
.include "../system/physics.h.s"
.include "../system/render.h.s"
.include "../system/ai.h.s"
.include "../system/collisions.h.s"
.include "../system/animations.h.s"
.include "../system/score.h.s"
.include "../system/input.h.s"
.include "../system/death_behaviours.h.s"
.include "../system/audio.h.s"
.include "../sprites/sprites.h.s"
.include "../templates/templates.h.s"

; Firmware
.globl cpct_disableFirmware_asm

; Random
.globl cpct_getRandom_xsp40_u8_asm

; Video
.globl cpct_setPALColour_asm

; Keyboard
.globl cpct_scanKeyboard_f_asm
.globl cpct_isKeyPressed_asm
.globl cpct_isAnyKeyPressed_asm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; GAME DATA ("VARIABLES")                                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Game data
gd_death_anim_counter::     .db 00
gd_data_flags::             .db 00
;; Cooldowns
gd_player_cooldown_cntr::   .db 10
;; Player data
gd_player_extra_lifes::     .db 03
gd_player_weapon_lvl::      .db 01

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Game Execute                                                     ;;
;;  Entry point, runs the game init and the game itself             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game:
    call man_game_init
    call man_game_run
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Game Init                                                        ;;
;;  Initializes the game                                            ;;
;; Input: -                                                         ;;
;; Output:  IX, AF, BC, DE and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_init:
    ;; Init render system
    call sys_render_game_init

    ;; Init game data
    call man_game_init_data

    ;; Init level manager
    call man_lvl_init

    ;; Init entities
    call man_game_starfield_init
    call man_game_respawn_player

    ;; Init music
    call sys_audio_init_game_music
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Game Init Data                                                   ;;
;;  Initializes the game data, such as player lifes                 ;;
;; Input: -                                                         ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_init_data:
    ;; Init extra lifes
    ld a, #player_extra_lifes
    ld (gd_player_extra_lifes), a

    ;; Set weapon level
    ld a, #player_starting_weapon
    ld (gd_player_weapon_lvl), a

    ;; Init death animation to Off
    ld a, #0
    ld (gd_death_anim_counter), a
    ;; Init extinction to Off
    ld (gd_data_flags), a

    ;; Init score to all zeros
    call sys_score_init

    ;; Init Basic HUD
    call man_game_init_hud
ret

man_game_init_hud:
    ;; Create HUD Background
    ld hl, #hud_template_background
    call man_create_one_hud_item

    ;; Init score string
    call sys_score_create

    ;; Draw life icons
    call man_game_init_extralife_counter
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Game Run                                                         ;;
;;  Runs the game's main loop                                       ;;
;; Input: -                                                         ;;
;; Output:  NZ and A!=0 if player wins, Z and A==0 if player loses  ;;
;;          Destroys IX, IY, AF, BC, DE and HL                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_run:
    main_loop:
        ;; Update level events
        call man_lvl_update_level

        ;; Update AI
        call sys_ai_update_all_entities

        ;; Update physics
        call sys_physics_update_all_entities

        ;; Update collisions
        call sys_collisions_update_all_entities

        ;; Update animations
        call sys_anim_update_all

        ;; Check if player wants to pause the game
        call sys_check_pause_game

        ;; Update render
        call sys_render_all

        ;; Update dead entities and hud items
        call sys_score_draw
        call man_update_dead_entities
        call man_update_dead_hud_items

        ;; Pause game if needed
        call sys_pause_game

        ;; Check extinctions
        call man_game_check_extinctions

        ;; Update game cooldowns and player's death animation
        call man_game_update_cooldown_counters
        call man_game_update_player_death_animation

        ;; Check if player wants to mute the game
        call sys_audio_check_mute

        ;; Check game beaten
        ld a, (gd_data_flags)
        and #0x80
            jp nz, jump_to_end_screen ;; If we beat the game, exit the loop

        ;;If player is dead, get out of the loop and init game end screen
        call man_game_check_death
            jp z, jump_to_end_screen    ;; If player ran out of extra lifes, exit game loop

        ;; Wait a frame
        call wait_a_frame

    jr main_loop

    ;; Loop exit point
    jump_to_end_screen:

    ;; Reset the screen buffer and entity array
    call sys_render_SWAP_to_C0
    call man_empty_entities
    call man_empty_hud

    ;; Hard screen erase
    cpctm_clearScreen_asm 0x00

    ;; Save Paused flag in B
    ld a, (gd_data_flags)
    and #0x04
    ld b, a

    ;; Set Z flag to 0 if player died, or to 1 if player beat the game
    ld a, (gd_data_flags)
    and #0x80   ;; A != 0 if A==0x80. So NZ (You win)
                ;; A == 0 if A!=0x80. So Z  (You died)

    ld a, #0 ;; To substract B later
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Cooldown Counters                                         ;;
;;  Update every game's cooldown                                    ;;
;; Input: -                                                         ;;
;; Output:  Destroys A and HL                                       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_update_cooldown_counters:
    ld hl, #gd_player_cooldown_cntr
    call man_game_update_cooldown_counter
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update One Cooldown                                              ;;
;;  Update every game's cooldown                                    ;;
;; Input: HL with cooldown counter meory position                   ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_update_cooldown_counter:
    ;; Check if cooldown is already zero
    ld a, (hl)
    or a
    ret z ;; If it's zero, return
        ;; If it's not, decrease
        dec a
        ld (hl), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Init Starfield                                                   ;;
;;  Initialize the starfield effect for the background              ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, AF, BC, DE and HL                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_starfield_init:
    .rept 3
    ld hl, #ent_template_star0
    call man_create_one_entity
    call man_game_positionate_random
    .endm

    .rept 4
    ld hl, #ent_template_star1
    call man_create_one_entity
    call man_game_positionate_random
    .endm

    .rept 5
    ld hl, #ent_template_star2
    call man_create_one_entity
    call man_game_positionate_random
    .endm

    .rept 5
    ld hl, #ent_template_star3
    call man_create_one_entity
    call man_game_positionate_random
    .endm
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Positionate Random                                               ;;
;;  Set an entity's position to random [0-38, 0-127]                ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys AF, DE and HL                                  ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_positionate_random:
    ;; Get a random X [0-31]
    call cpct_getRandom_xsp40_u8_asm
    ld a, l
    ld b, #31
    and b
    ld ent_x_pos(ix), a

    ;; Get a random X [0-7]
    call cpct_getRandom_xsp40_u8_asm
    ld a, l
    ld b, #7
    and b

    ;; Add both Xs (31+7 -> [0-38])
    ld b, ent_x_pos(ix)
    add b
    ld ent_x_pos(ix), a

    ;; Get a random Y [0-127]
    call cpct_getRandom_xsp40_u8_asm
    ld a, l
    ld d, #127
    and d
    ld ent_y_pos(ix), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Respawn Player                                                   ;;
;;  Respawns a player entity, if no other player is alive           ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_respawn_player:
    ;; Check if player is alive. Z=0 if it is
    call man_game_check_player_status

    jr nz, respawn_player_skip ;; If A!=0, another player is alive, so skip respawn
        ;; Create a new player
        ld hl, #ent_template_player
        call man_create_one_entity

    respawn_player_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check if Player is alive                                         ;;
;;  Checks if player is alive                                       ;;
;; Input: -                                                         ;;
;; Output:  Destroys B, A // Return 0 in a if dead, or 1 if alive   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_check_player_status:
    ld b, #ent_type_player  ;; Load player mask
    call man_search_type    ;; Search for players
    or a                   ;; A = 0 if dead, so Z = 1 if no player was found
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Death of Player                                            ;;
;;  Checks if player is dead and decreases the life count if is dead;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_check_death:
    ;; Check status. Z=1 if player is dead / no player is alive
    call man_game_check_player_status
    jr nz, player_is_alive  ;; If Not Z, skip life decrease

        ;; Decrease life count
        ld a, (gd_player_extra_lifes)
        dec a
        ld (gd_player_extra_lifes), a

        ;; If life was zero before decreaes (now is negative / S=1) then player is fully dead
        jp m, player_is_alive
            ;; Clean all entities and screen
            call man_empty_entities
            call man_empty_hud
            cpctm_clearScreen_asm 0x00

            ;; Reinit starfield
            call man_game_starfield_init

            ;; Respawn player if it is alive
            call man_game_respawn_player

            ;; Decrease weapon level
            ld a, (gd_player_weapon_lvl)
            dec a
            cp #1
            jr nc, player_weapon_decrease_adjust_skip
                ld a, #1
            player_weapon_decrease_adjust_skip:
            ld (gd_player_weapon_lvl), a
            call man_set_player_attack_pattern

            ;; Load last check point
            call man_lvl_load_last_checkpoint

            ;; Redraw HUD
            call man_game_init_hud

            ;; Reset Hi Mark
            call check_hi_score_reached

            ;; Restart graphic configuration
            call sys_render_game_init
            ld a, #0
            ld (gd_death_anim_counter), a

    player_is_alive:
    call man_game_check_player_status ;; Set Z flag to 1 if player is dead
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Extinctions                                                ;;
;;  Checks if someone must go extinct                               ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_check_extinctions:
    call man_game_check_enemy_extinction
    call man_game_check_player_extinction
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Enemy Bomb                                                 ;;
;;  Checks if player must die                                       ;;
;;  (can be triggered by "Grab the Flag" bosses)                    ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_check_player_extinction:
    ;; Check extinction (!= 0)
    ld a, (gd_data_flags)
    and #0x02 ;; Bit 1
    ret z   ;; If it's zero, do nothing

        ;; Save a zero again (to avoid extinction on next frame)
        ld a, (gd_data_flags)
        xor #0x02
        ld (gd_data_flags), a

        ;; Mark player for death
        ld b, #ent_type_player
        call man_mark_for_destroy_entity_of_type
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Player Bomb                                                ;;
;;  Checks if all enemies must die                                  ;;
;;  (can be triggered by defeated bosses)                           ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_check_enemy_extinction:
    ;; Check extinction (!= 0)
    ld a, (gd_data_flags)
    and #0x01 ;; Bit 0
    ret z   ;; If it's zero, do nothing

        ;; Save a zero again (to avoid extinction on next frame)
        ld a, (gd_data_flags)
        xor #0x01
        ld (gd_data_flags), a

        ;; Mark all enemies and their bullets for death
        ld b, #(ent_type_enemy | ent_type_enemy_bullet)
        call man_mark_for_destroy_entity_of_type
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Activate Bomb                                                    ;;
;;  Makes player or all enemies to die in next frame                ;;
;;  (can be triggered by defeated bosses)                           ;;
;; Input: -                                                         ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_activate_enemy_extinction:
    ;; Set extinction flag's bit 0 to 1
    ld a, (gd_data_flags)
    or #1
    ld (gd_data_flags), a
ret
man_game_activate_player_extinction:
    ;; Set extinction flag's bit 1 to 1
    ld a, (gd_data_flags)
    or #2
    ld (gd_data_flags), a
ret
man_game_mark_as_beaten:
    ;; Set game beaten's bit flag to 1
    ld a, (gd_data_flags)
    or #0x80
    ld (gd_data_flags), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Creates Extra life HUD                                           ;;
;;  Creates a hud item to show the player their extra life count    ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_init_extralife_counter:
    ;; Create ShipX icon
    ld hl, #hud_template_shipx_icon
    call man_create_one_hud_item

    call man_game_create_extralife_counter
ret

man_game_create_extralife_counter:
    ;; Create number element
    ld hl, #hud_template_extralife_number   ;; Template of extra life counter
    call man_create_one_hud_item
    
    ld a, (gd_player_extra_lifes)           ;; Value to show
    call sys_score_get_number_sprite        ;; Change sprite to lifes number
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Add Extra Life                                                   ;;
;;  Adds an extra life to the player and updates the hud            ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_add_extralife:
    ;; Decrease internal counter
    ld hl, #gd_player_extra_lifes
    inc (hl)

    ;; Check max life count
    ld b, (hl)                      ;; B = current lifes (increased)
    ld a, #player_max_extra_lifes   ;; A = max lifes
    sub b                           ;; Max - Current (must be >= 0)
    jp p, skip_life_cap            ;; If max-current < 0, then Current > Max so Current is capped at Max

        ;; Correct current level to max level
        ld a, #player_max_extra_lifes
        ld (gd_player_extra_lifes), a

        ;; Give points
        call sys_death_earn_points

    skip_life_cap:

    ;; Destroy HUD life counter
    ld b, #hud_type_lifebar
    call man_mark_for_destroy_hud_of_type

    ;; Redraw HUD life counter
    call man_game_create_extralife_counter
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Upgrade Weapon                                                   ;;
;;  Adds a level to the player's attack pattern                     ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, B, DE and HL                            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_weapon_up:
    ;; Increase weapon level
    ld hl, #gd_player_weapon_lvl    ;; A = current weapon level
    inc (hl)                        ;; Increase weapon level

    ;; Check max level
    ld b, (hl)                      ;; B = current level (increased)
    ld a, #player_max_weapon_lvl    ;; A = max level
    sub b                           ;; Max - Current (must be >= 0)
    jp p, skip_level_cap            ;; If max-current < 0, then Current > Max so Current is capped at Max

        ;; Correct current level to max level
        ld a, #player_max_weapon_lvl
        ld (gd_player_weapon_lvl), a

        ;; Give points
        call sys_death_earn_points

    skip_level_cap:

    call man_set_player_attack_pattern
ret

man_set_player_attack_pattern:
    ;; Get player's memory pos in IX
    ld b, #ent_type_player
    call man_search_type    ;; IX = player memory

    ;; Get current level
 switch_wpn_lvl_size = 2 + 9
    ld a, (gd_player_weapon_lvl)    ;; A = current weapon level
    dec a
    jr nz, .+switch_wpn_lvl_size            ;; 2 bytes
        ;; Load attack pattern 1 into player
        ld hl, #man_bullet_player_lvl1  ;; 3 bytes
        ld ent_atk_ptrn(ix)  , l        ;; 3 bytes
        ld 1+ent_atk_ptrn(ix), h        ;; 3 bytes
    dec a
    jr nz, .+switch_wpn_lvl_size
        ;; Load attack pattern 2 into player
        ld hl, #man_bullet_player_lvl2
        ld ent_atk_ptrn(ix)  , l    
        ld 1+ent_atk_ptrn(ix), h 
    dec a
    jr nz, .+switch_wpn_lvl_size
        ;; Load attack pattern 3 into player
        ld hl, #man_bullet_player_lvl3
        ld ent_atk_ptrn(ix)  , l    
        ld 1+ent_atk_ptrn(ix), h 
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Shoot                                                            ;;
;;  Checks if we can create new bullet                              ;;
;; Input: -                                                         ;;
;; Output:  Destroys A, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_shoot:
    ;; Save shooter entity
    push iy

    ;; Check if player can shoot
    ld a, (gd_player_cooldown_cntr)
    or a
    jr nz, no_shoot  ;; If cooldown > 0, player can't shoot

        ;; Shooting SFX
        call sys_audio_play_sfx_playershoot

        ;; Reset cooldown to avoid multiple shooting
        ld a, #player_shoot_cooldown
        ld (gd_player_cooldown_cntr), a

        call man_bullet_shoot_bullet

    no_shoot:

    ;; Recover shooter entity
    pop iy
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Start Death Animation                                            ;;
;;  Enables the deth screen animation update                        ;;
;; Input: -                                                         ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_start_player_death_animation:
    ld a, #0x80
    ld (gd_death_anim_counter), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Death Animation                                           ;;
;;  Checks if player died and makes a death animation modifying     ;;
;;  screen properties and palette colours                           ;;
;; Input: -                                                         ;;
;; Output:  Destroys AF, B, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_game_update_player_death_animation:
    ;; Update the counter and check if we're dead
    ld a, (gd_death_anim_counter)   ;; Load counter
    rrca                            ;; Move the bit
    ld (gd_death_anim_counter), a   ;; Save counter

    or a                           ;; Check if A = 0
    jp z, skip_death_update         ;; If number is still zero we're not dying, so skip this routine

        ;; Save counter backup in D
        ld d, a

        ;; Select animation frame to execute
        and #0b01000000 ;; Bit 7
        jr z, skip_bit_7
            ;; BIT 7 FRAME

            ;; Set border to red
            ld l, #0x10                 ;; Select border
            ld h, #0x1C                 ;; Red border
            call cpct_setPALColour_asm  ;; Input: HL, destroys AF and B

            ;; Set background to red
            ld l, #0x00                 ;; Select background
            ld h, #0x0C                 ;; Light Red
            call cpct_setPALColour_asm  ;; Input: HL, destroys AF and B

            ;; Modify HSync
            ld de, #0x0223
            call sys_render_change_CRTC

            jp skip_death_update        ;; Jump to end
        skip_bit_7:

        ;ld a, d
        ;and #0b00100000 ;; Bit 6
        ;jr z, skip_bit_6
        ;    ;; BIT 6 FRAME
        ;    jp skip_death_update        ;; Jump to end
        ;skip_bit_6:

        ;ld a, d
        ;and #0b00010000 ;; Bit 5
        ;jr z, skip_bit_5
        ;    ;; BIT 5 FRAME
        ;    jp skip_death_update        ;; Jump to end
        ;skip_bit_5:

        ;ld a, d
        ;and #0b00001000 ;; Bit 4
        ;jr z, skip_bit_4
        ;    ;; BIT 4 FRAME
        ;    jp skip_death_update        ;; Jump to end
        ;skip_bit_4:

        ld a, d
        and #0b00000100 ;; Bit 3
        jr z, skip_bit_3
            ;; BIT 3 FRAME

            ;; Set border to light red
            ld l, #0x10                 ;; Select border
            ld h, #0x0E                 ;; Orange border
            call cpct_setPALColour_asm  ;; Input: HL, destroys AF and B

            ;; Set background to red
            ld l, #0x00                 ;; Select background
            ld h, #0x1C                 ;; Red
            call cpct_setPALColour_asm  ;; Input: HL, destroys AF and B

            ;; Modify HSync
            ld de, #0x0225
            call sys_render_change_CRTC

            jp skip_death_update        ;; Jump to end
        skip_bit_3:

        ;ld a, d
        ;and #0b00000010 ;; Bit 2
        ;jr z, skip_bit_2
        ;    ;; BIT 2 FRAME
        ;    jp skip_death_update        ;; Jump to end
        ;skip_bit_2:

        ;ld a, d
        ;and #0b00000001 ;; Bit 1
        ;jr z, skip_bit_1
        ;    ;; BIT 1 FRAME
        ;    jp skip_death_update        ;; Jump to end
        ;skip_bit_1:

    skip_death_update:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Pause Game                                                 ;;
;; Check if we need to pause the game                               ;;
;; Input: -                                                         ;;
;; Output:  Destroys AF, BC, DE, IX and HL                          ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_check_pause_game:

    ld hl, #keyid_x
    call cpct_isKeyPressed_asm
    ret z

        ld hl, #hud_template_pause_msg
        call man_create_one_hud_item

        ld a, (gd_data_flags)
        or #0x04
        ld (gd_data_flags), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Pause Game                                                       ;;
;; Check if we need to pause the game                               ;;
;; Input: -                                                         ;;
;; Output:  Destroys AF, BC, DE and HL                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_pause_game:
    ld a ,(gd_data_flags)
    and #0x04
    ret z 
        pause_multiple_keypress_check:
            call cpct_scanKeyboard_f_asm
            ld hl, #keyid_x                  ;;If Resume key is pressed
            call cpct_isKeyPressed_asm  
        jr nz, pause_multiple_keypress_check

        pause_loop::
            call cpct_scanKeyboard_f_asm

            ld hl, #keyid_x                  ;;If Resume key is pressed
            call cpct_isKeyPressed_asm  
            jp nz , is_x_pressed            ;;Wait until is released

            ld hl, #keyid_esc                ;;If Esc key is pressed
            call cpct_isKeyPressed_asm 
            jp nz , is_esc_pressed          ;; Wait ultil is released

        jp pause_loop

        is_x_pressed::

            call cpct_scanKeyboard_f_asm
            call cpct_isAnyKeyPressed_asm
            jp nz, is_x_pressed

            ld a,(gd_data_flags)            ;; Deactive pause flag
            and #0xFF - 4
            ld (gd_data_flags), a           ;; Save new state of the flag

            ld b, #hud_type_pause           ;; Search Pause Sprite
            call man_hud_search_type

            ;; Roll prev_render to prev_prev_render
            ld l, hud_prev_render(ix)
            ld h, 1+hud_prev_render(ix)
            ld hud_prev_prev_render(ix), l
            ld 1+hud_prev_prev_render(ix), h

            ld b, #hud_type_pause           ;; Destroy Pause Sprite
            call man_mark_for_destroy_hud_of_type

        ret 

        is_esc_pressed::

            call cpct_scanKeyboard_f_asm
            call cpct_isAnyKeyPressed_asm
            jp nz, is_esc_pressed

            ld a,(gd_data_flags)            ;; Deactive pause flag
            or #0x80
            ld (gd_data_flags), a

            ret

            
ret