;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "huditems.h.s"
.include "../sprites/sprites.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Variables                                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
huditems::          .ds hud_item_size*hud_max_count
hud_array_end:      .db 0x00
hud_array_free_space:         .db hud_max_count
hud_array_next_free_position: .dw #huditems

.globl cpct_memset_asm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Hud Item                                                  ;;
;;  Creates an hud item using the specified template                ;;
;; Input: HL with the template to use (memory position)             ;;
;; Output:  IX with new hud item's mem pos, A = 0 if no new hud item;;
;;          Destroys BC and DE                                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_create_one_hud_item:
    ;; Check if there is free space left
    ld a, (hud_array_free_space)
    or a
    jr z, man_create_one_hud_skip

    ;; Search next free position
    ld ix, (hud_array_next_free_position)

    ;; Load template on new hud item
    ;; HL points to template array
    push ix                     ;; Use the pile to pass data to DE
    pop de                      ;; DE points to the start of the new hud item
    ld bc, #hud_item_size       ;; BC holds bytes to copy
    ;; HL                       ;; HL holds the start of the template
    ldir

    ;; Add "Draw next time" bit so it's rendered the first time at least
    ld a, hud_comps(ix)
    or #hud_comp_redraw1
    or #hud_comp_redraw2
    or #hud_comp_redraw3
    ld hud_comps(ix), a

    ;; Change next free position to [hud item size] onwards
    ;; Also leave IX pointing to new hud item
    push ix                 ;; Save new hud item's position
    ld bc, #hud_item_size   ;; Load the offset to reach next hud item's start
    add ix, bc              ;; Add it to IX
    ld (hud_array_next_free_position), ix ;; Save it
    pop ix                  ;; Recover new hud item's position

    ;; Decrease free space counter
    ld a, (hud_array_free_space)
    dec a
    ld (hud_array_free_space), a
    
    ld a, #0x01 ;; Load A output (= 1 if new hud item was created)
    jr man_create_one_hud_end

    ;; Label for creation skipping if there is no free space left
    man_create_one_hud_skip:
    ld a, #0x00 ;; Load A output (= 0 if NO new hud item was created)

    man_create_one_hud_end:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Destroy Hud item                                                 ;;
;;  Destroys the specified hud item, and reorganizes the array      ;;
;; Input: IX with hud item's memory position                        ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_destroy_one_hud_item:
    ;; Load HL with last hud item's memory position
    ld hl, (hud_array_next_free_position);; HL stores the origin
    ld a, #hud_item_size
    destroy_one_go_back_an_hud_item:
        dec hl
        dec a
    jr nz, destroy_one_go_back_an_hud_item

    ;; Save next free position
    ld (hud_array_next_free_position), hl

    ;; Copy last hud item into dead hud item
    push ix
    pop de                                  ;; DE must store the destiny (Dead hud item)
    ld bc, #hud_item_size                     ;; BC holds the number of bytes to copy
    ;; HL                                   ;; HL holds the origin of the hud item to copy
    ldir

    ;; Mark the moved hud item's old position as free
    ld hl, (hud_array_next_free_position) 
    ld (hl), #0x00 ;; Mark as invalid  

    ;; Update hud item's array free space
    ld hl, #hud_array_free_space
    inc (hl)
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mark Hud item for Destruction                                    ;;
;;  Marks the specified hud item for later destruction              ;;
;; Input: IX with hud item's memory position                        ;;
;; Output: Destroys A                                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_mark_hud_item_for_destruction:
    ;; Mark a hud item as dead using the dead flag
    ld a, hud_type(ix)  ;; Load the type
    or #hud_type_dead   ;; Add death mask bit to A
    ld hud_type(ix), a  ;; Save that on type
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mark Hud item for Destruction of Type                            ;;
;;  Marks all hud items of the specified type for later destruction ;;
;; Input: B with filter mask                                        ;;
;; Output: Destroys IX, A and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_mark_for_destroy_hud_of_type:
    ;; Load hud array start
    ld ix, #huditems

    ;; Loop until it finds an invalid hud item
    man_hud_mark_type_loop:
        ;; Check if entity is valid
        ld a, hud_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        jr z, man_hud_mark_type_skip ;; If its zero (invalid) finsih te loop

        ;; Check if is one of the types we want to destroy
        ;;ld a, hud_type(ix)    ;; A = type
        and b                   ;; B = filter mask
        jr z, mark_type_destruction_skip   ;; If zero, no filter passed, skip death

        ;; If its valid, destroy it
        call man_mark_hud_item_for_destruction
        mark_type_destruction_skip:

        ;; Move to next hud item
        ld de, #hud_item_size
        add ix, de

    ;; Loop end
    jr man_hud_mark_type_loop

    man_hud_mark_type_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Dead Entites                                              ;;
;;  Search the array for marked-as-dead entites to destroy them     ;;
;;  at once                                                         ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B and DE                                 ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_update_dead_hud_items:
    ;; Load hud item array start
    ld ix, #huditems

    ;; Loop until it finds an invalid hud item
    man_update_hud_loop:
        ;; Check if hud item is valid
        ld a, hud_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        jr z, man_update_hud_skip ;; If its zero (invalid) finsih te loop

        ;; Check if hud item is marked for death
        ;; A still stores the type
        ld b, #hud_type_dead;; Load dead mask
        and b               ;; Bitwise AND with death mask is Zero if death flag is Zero (not dead)
        jr z, man_update_hud_death_skip

        ;; Kill the hud item (IX is the input)
        call man_destroy_one_hud_item
        jr man_update_hud_increase_skip    ;; If we destroyed one, we do not advance to the next position
        
        ;; Move to next hud item
        man_update_hud_death_skip: ;; Label for death skip
        ld de, #hud_item_size
        add ix, de

        man_update_hud_increase_skip:

    ;; Loop end
    jr man_update_hud_loop

    man_update_hud_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Apply to All Hud Items                                           ;;
;;  Apply the specified function to every hud item                  ;;
;; Input: HL with routine's memory pos                              ;;
;; Output: Destroys IX, A and DE. Might destroy F and BC            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_hud_for_all:
    ;; Load hud array start
    ld ix, #huditems

    ;; Loop until it finds an invalid hud item
    man_hud_for_all_loop:
        ;; Check if entity is valid
        ld a, hud_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        jr z, man_hud_for_all_skip ;; If its zero (invalid) finsih te loop

        ;; Save input parameters data
        push hl

        ;; If its valid, apply routine
        ld de, #man_hud_for_all_return_point    ;; Load return position
        push de                                 ;; Push that position to the pile
        jp (hl)                                 ;; CALL simulated with manual push and jump

        ;; Load input parameters data
        man_hud_for_all_return_point:
        pop hl

        ;; Move to next hud item
        ld de, #hud_item_size
        add ix, de

    ;; Loop end
    jr man_hud_for_all_loop

    man_hud_for_all_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Search a Matching Hud Item                                       ;;
;;  Search for an specific hud item type                            ;;
;; Input: B with the filter mask                                    ;;
;; Output: A=z=1 if found, A=z=0 if not.  Destroys IX and DE       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_hud_search_type:
    ;; Load entity array start
    ld ix, #huditems

    ;; Loop until it finds an invalid entity
    man_hud_search_type_loop:
        ;; Check if entity is valid
        ld a, hud_type(ix)  ;; Load the type in A
        or a                ;; Check zero value
        jr z, man_hud_search_not_found ;; If its zero (invalid) finsih te loop

        ;; Check if entity has the desired components
        ;; A = type mask    ;; A stores type mask
        and b               ;; Bitwise AND with type filter mask
        jr nz, man_hud_search_found_it

        ;; Move to next entity
        ld de, #hud_item_size
        add ix, de

    ;; Loop end
    jr man_hud_search_type_loop

    ;; Label for succesful search
    man_hud_search_found_it:
    ld a, #0x01
    jr man_hud_search_end

    ;; Label for search fail, array loop finished
    man_hud_search_not_found:
    ld a, #0x00

    man_hud_search_end:
    or a
    ;; Zero register is 0 if not found
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Empty the hud array                                              ;;
;;  Sets the hud array to 00                                        ;;
;; Input: -                                                         ;;
;; Output: Destroys A, BC and DE                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_empty_hud:
    ;; load in DE the start of the entities array
    ld de, #huditems
    ;; load in A 0 to fill the array
    ld a, #0x00
    ;; load in BC the size in BYTES
    ld bc, #hud_array_byte_size
    ;; do the same as in DE
    ld hl, #huditems
    ;;the next free hud item is the begining of the array yay
    ld (hud_array_next_free_position), hl
    ;;fill array with 0s
    call cpct_memset_asm
    ;;load in A the max entities cap
    ld a, #hud_max_count
    ;;set the free space to the max count
    ld (hud_array_free_space), a
ret
