;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "level_reader.h.s"
.include "game.h.s"
.include "../levels/levels.h.s"
.include "../manager/entities.h.s"
.include "../manager/huditems.h.s"
.include "../system/score.h.s"
.include "../system/audio.h.s"
.include "../templates/templates.h.s"
.include "cpctelera.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DATA                                                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
lvl_reader_pointer::      .dw 0x0000
lvl_reader_counter::      .db 1
lvl_reader_checkpoint::   .dw 0x0000
lvl_reader_keep_reading:: .db 1

;; Array to know the size of each event type
lvl_reader_evt_size_array::
;;  Wait | Enemy | Pow | Title | Finish | Checkp | Stop | Boss
.db  2,      4,     4,     3,      2,       2,      2,      3

;; Array to select enemy spawn
lvl_reader_enemy_select::
.dw #ent_template_enemy             ;;  1 = Dumb enemy
.dw #ent_template_enemy2            ;;  2 = Kamikaze
.dw #ent_template_enemy_shooty      ;;  3 = Dumb enemy that shoots
.dw #ent_template_enemy_zig         ;;  4 = Zigzagger
.dw #ent_template_enemy_patroller   ;;  5 = Patroller (keeps up moving in X and shooting)
.dw #ent_template_enemy_crawler     ;;  6 = Crawls through the screen (culebrilla)
.dw #ent_template_enemy_crawler_sh  ;;  7 = Crawler that shoots when it reaches a border
.dw #ent_template_enemy_barrier     ;;  8 = Barrier (Invincible)
.dw #ent_template_enemy_stalker     ;;  9 = Stalker (Follows you on X axis while going down)
;; BOSSES
.dw #ent_template_enemy_core        ;; 10 = Core (Stays up quiet, doing nothing, kills all enemies upon death)
.dw #ent_template_enemy_grabflag    ;; 11 = Grab the Flag (Goes down, dies on "pickup", but kills the player upon reaching the bottom)
    enemy_selec_array_size = ((. - lvl_reader_enemy_select)/2)

;; Array to select powerup spawn
lvl_reader_powerup_select::
.dw #ent_template_pw_extra_life     ;;  1 = Extra life
.dw #ent_template_pw_upgrade        ;;  2 = Weapon upgrade
.dw #ent_template_pw_points         ;;  3 = Earn Points
.dw #ent_template_pw_bomb           ;;  4 = Bomb (Delete all enemies)
.dw #ent_template_pw_persia         ;;  5 = Desert Hourglass (2500 pts)
    powerup_selec_array_size = ((. - lvl_reader_powerup_select)/2)

;; Array to select title spawn
lvl_reader_title_select::
.dw #hud_template_level1_title      ;;  1 = Title of level 1
.dw #hud_template_level2_title      ;;  2 = Title of level 2
.dw #hud_template_level3_title      ;;  3 = Title of level 3
.dw #hud_template_level4_title      ;;  4 = Title of level 4
.dw #hud_template_level5_title      ;;  5 = Title of level 5
.dw #hud_template_level6_title      ;;  6 = Title of level 6
.dw #hud_template_level7_title      ;;  7 = Title of level 7
.dw #hud_template_level8_title      ;;  8 = Title of level 8
.dw #hud_template_level9_title      ;;  9 = Title of level 9
.dw #hud_template_level10_title     ;; 10 = Title of level 10
.dw #hud_template_level11_title     ;; 11 = Title of level 11
.dw #hud_template_level12_title     ;; 12 = Title of level 12
.dw #hud_template_level13_title     ;; 13 = Title of level 13
.dw #hud_template_warning_title     ;; 14 = Warning title
    title_selec_array_size = ((. - lvl_reader_title_select)/2)

;; Array to select boss title spawn
lvl_reader_bosstitle_select::
.dw #hud_template_bosscoremsg       ;;  1 = Kill the Core
.dw #hud_template_bosscellmsg       ;;  2 = Grab the Power Cell
    bosstitle_selec_array_size = ((. - lvl_reader_bosstitle_select)/2)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Init level reader                                                ;;
;;  Load first level and init other data                            ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX and A                                       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_init:
    ;; Load first level's data into our reading pointer
    ld ix, #levels_script
    ld (lvl_reader_pointer), ix

    ;; Reset reader counter to level's first event
    ld a, evt_time(ix)
    ld (lvl_reader_counter), a

    ;; Init checkpoint to start of first level
    ld (lvl_reader_checkpoint), ix

    ;; Init "keep reading" flag to read
    ld a, #1
    ld (lvl_reader_keep_reading), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update level                                                     ;;
;;  Checks event counter and executes next step if necessary        ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_update_level:
    ;; Check if level reading is stopped and must be continued
    call man_lvl_check_continue
    ret z   ;; If returns Z, skip level reading

    ;; Check update counter
    ld a, (lvl_reader_counter)  ;; Get counter
    dec a                       ;; Decrease counter
    ld (lvl_reader_counter), a  ;; Save new counter value
    ret nz                      ;; If zero, load next level event
        ;; Execute current event
        ld ix, (lvl_reader_pointer)
        event_execution:
        call man_lvl_manage_event ;; Redirection: This routine redirects to multiple routines

        ;; Load next event
        ld ix, (lvl_reader_pointer) ;; Load IX backup
        call man_lvl_get_evt_size   ;; BC = Event size
        add ix, bc                  ;; IX = IX + BC = Old_event + event_size = New_event
        ld (lvl_reader_pointer), ix ;; Save new event's pointer

        ;; Check if time is zero to execute next event in this cycle
        ld a, evt_time(ix)
        or a    ;; Check if A (event time) is zero
        jr z, event_execution

        ;; Load new event's counter
        ld a, evt_time(ix)
        ld (lvl_reader_counter), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Load Last Checkpoint                                             ;;
;;  Load checkpoint into level reader to return to the checkpoint   ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX                                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_load_last_checkpoint:
    ;; Load checkpoint
    ld ix, (lvl_reader_checkpoint)
    ld (lvl_reader_pointer), ix

    ;; Load time of new first event
    ld a, evt_time(ix)
    ld (lvl_reader_counter), a

    ;; Load previous score
    call sys_score_load_score_state
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Continue                                                   ;;
;;  Checks if level is stopped, and tries to continue it            ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, B and DE                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_check_continue:
    ;; Check if level reading is enabled (can be stopped and continued by events)
    ld a, (lvl_reader_keep_reading)
    or a
    ret nz      ;; If NZ, go back to level reading

        ;; Check if enemies are gone
        ld b, #ent_type_enemy
        call man_search_type
        jr nz, level_continue_check_skip ;; If we find enemies, do nothing

            ;; Re-enable reading
            call man_lvl_continue_reading
            ;; Return NZ
            ld a, #1
            or a
            ;; NZ
            ret
        level_continue_check_skip:
        ;; Return Z
        ld a, #0
        or a
        ;; Z
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Continue Level Reading                                           ;;
;;  Mark the level reader to continue reading                       ;;
;; Input: -                                                         ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_continue_reading:
    ;; Init "keep reading" flag to read
    ld a, #1
    ld (lvl_reader_keep_reading), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Get Event Size                                                   ;;
;;  Loads the current event size in BC (each event type has a size) ;;
;; Input: -                                                         ;;
;; Output:  BC with event size.  Destroys HL and A                  ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_get_evt_size:
    ;; Load previous byte to size array
    ld hl, #lvl_reader_evt_size_array
    dec hl

    ;; Load event type to be used as a counter (to search the array)
    ld a, evt_type(ix)

    ;; Search the array for the specified event type
    evt_size_loop:
        inc hl  ;; Advance HL to next position of array
        dec a   ;; Decrease counter
    jr nz, evt_size_loop    ;; Repeat until counter == 0

    ;; HL holds memory position of event type's size
    ld a, (hl)  ;; A = event size (1 byte)
    ld b, #0    ;; | BC = #0x00ss (ss = size byte = A) = Event size (2 bytes)
    ld c, a     ;; |
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manage Event                                                     ;;
;;  Redirectionates the event management to its respective type     ;;
;;  manager                                                         ;;
;; Input: IX with event data pointer                                ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_manage_event:
    ;; Load event type
    ld a, evt_type(ix)

    dec a
    ret z                           ;; Type 1 manager (Do nothing)
    dec a
    jp z, #man_lvl_spawn_enemy      ;; Type 2 manager (Create enemy)
    dec a
    jp z, #man_lvl_spawn_powerup    ;; Type 3 manager (Create powerup)
    dec a
    jp z, #man_lvl_toggle_title     ;; Type 4 manager (Show or Hide title)
    dec a
    jp z, #man_lvl_finish           ;; Type 5 manager (Finish game)
    dec a
    jp z, #man_lvl_save_checkpoint  ;; Type 6 manager (Save checkpoint)
    dec a
    jp z, #man_lvl_stop_reading     ;; Type 7 manager (Stop reading events)
    dec a
    jp z, #man_lvl_toggle_boss_title;; Type 8 manager (Show or Hide boss title)
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 2 Manager - Create an enemy                                 ;;
;;  Creates an enemy and set its X to random                        ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, IY, A, BC, DE and HL                       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_spawn_enemy:
    ld a, evt_param1(ix) ;; Param = Enemy type / Enemy Template

    ;; Save event data in IY to use it later
    push ix
    pop iy

    ;; Check if parameter is outside array size
    ld b, #enemy_selec_array_size   ;; B = array size
    inc b                           ;; B = array size + 1
    cp b                            ;; A - B+1 < 0 if A<=B (A is in range)
    jr c, enemy_param_fix_skip
        ld a, #1
    enemy_param_fix_skip:

    ;; Select enemy template (Load in HL)
    ld hl, #lvl_reader_enemy_select
    dec hl
    dec hl
    enemy_selection_loop:
        inc hl
        inc hl
        dec a
    jr nz, enemy_selection_loop

    ;; Load (HL) in HL
    ld c, (hl)
    inc hl
    ld b, (hl)
    ld h, b     ;; HL = Enemy template memory
    ld l, c

    ;; Create the enemy from the selected template (stored in HL)
    call man_create_one_entity

    ;; Check if new entity was created
    or a
    jr z, enemy_spawn_skip ;; If no enemy was created, skip

        ;; Modify its X position
        ld a, evt_param2(iy) ;; Get desired X position
        ld b, ent_spr_w(ix)  ;; Get entity's width
        srl b                ;; B = Half width
        sub b                ;; A = Xpos - W/2
        ld ent_x_pos(ix), a

    enemy_spawn_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 3 Manager - Create a powerup                                ;;
;;  Creates a powerup and set its X to random                       ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, IY, A, BC, DE and HL                       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_spawn_powerup:
    ld a, evt_param1(ix) ;; Param = Powerup type / Powerup Template

    ;; Save event data in IY to use it later
    push ix
    pop iy

    ;; Check if parameter is outside array size
    ld b, #powerup_selec_array_size   ;; B = array size
    inc b                           ;; B = array size + 1
    cp b                            ;; A - B+1 < 0 if A<=B (A is in range)
    jr c, powerup_param_fix_skip
        ld a, #1
    powerup_param_fix_skip:

    ;; Select powerup template (Load in HL)
    ld hl, #lvl_reader_powerup_select
    dec hl
    dec hl
    powerup_selection_loop:
        inc hl
        inc hl
        dec a
    jr nz, powerup_selection_loop

    ;; Load (HL) in HL
    ld c, (hl)
    inc hl
    ld b, (hl)
    ld h, b     ;; HL = Powerup template memory
    ld l, c

    ;; Create the powerup from the selected template (stored in HL)
    call man_create_one_entity

    ;; Check if new entity was created
    or a
    jr z, powerup_spawn_skip ;; If no powerup was created, skip
        
        ;; Modify its X position
        ld a, evt_param2(iy)
        ld b, ent_spr_w(ix)  ;; Get entity's width
        srl b                ;; B = Half width
        sub b                ;; A = Xpos - W/2
        ld ent_x_pos(ix), a

    powerup_spawn_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 4 Manager - Show or Hide screen string                      ;;
;;  Change the screen string to the specified sprite, and show it or;;
;;  hide it                                                         ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_toggle_title:
    ;; Check if a new title must be added, or the current title must be destroyed
    ld b, #hud_type_title       ;; Load title filter
    push ix                         ;; Save IX backup
    call man_hud_search_type    ;; Search for titles
    pop ix                          ;; Load IX backup
    jr z, spawn_new_title       ;; If zero, no title found. A new title must be created

        ;; Delete title
        ld b, #hud_type_title
        call man_mark_for_destroy_hud_of_type
        jp title_spawn_skip

    spawn_new_title:
        ;; Select title template
        ld a, evt_param1(ix) ;; Param = Title to show (With sprite, anim, width height, xpos...)

        ;; Check if parameter is outside array size
        ld b, #title_selec_array_size   ;; B = array size
        inc b                           ;; B = array size + 1
        cp b                            ;; A - B+1 < 0 if A<=B (A is in range)
        jr c, title_param_fix_skip
            ld a, #1
        title_param_fix_skip:

        ;; Select title template (Load in HL)
        ld hl, #lvl_reader_title_select
        dec hl
        dec hl
        title_selection_loop:
            inc hl
            inc hl
            dec a
        jr nz, title_selection_loop

        ;; Load (HL) in HL
        ld c, (hl)
        inc hl
        ld b, (hl)
        ld h, b     ;; HL = Title template memory
        ld l, c

        ;; Create the title
        call man_create_one_hud_item
    title_spawn_skip:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 5 Manager - Finish Game                                     ;;
;;  Mark the game as finished to win the game                       ;;
;; Input: -                                                         ;;
;; Output:  Destroy A and HL                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_finish:
    ;; Create the AI player that escapes
    ld b, #ent_type_player
    call man_search_type ;; A,IX,DE
    jp z, man_lvl_finish_player_not_found ;; If not found, beat game without animation
        ;; Save player position and Erasing position
        ld h, ent_x_pos(ix)
        ld l, ent_y_pos(ix)
        push hl
        ld l, ent_prev_render(ix)
        ld h, 1+ent_prev_render(ix)
        push hl
        ld l, ent_prev_prev_render(ix)
        ld h, 1+ent_prev_prev_render(ix)
        push hl

        ;; Swap player with an AI-controlled version that escapes
        push ix                             ;; Use the pile to pass data to DE
        pop de                              ;; DE points to the start of the new entity
        ld bc, #entity_size                 ;; BC holds bytes to copy
        ld hl, #ent_template_player_escape  ;; HL holds the start of the template
        ldir                                ;; Copy HL into DE (BC bytes)

        ;; Load position and video memory
        pop hl
        ld ent_prev_prev_render(ix)  , l
        ld 1+ent_prev_prev_render(ix), h
        pop hl
        ld ent_prev_render(ix)  , l
        ld 1+ent_prev_render(ix), h
        pop hl
        ld ent_x_pos(ix), h
        ld ent_y_pos(ix), l

        ;; Throw win music
        call sys_audio_init_win_music

        ret
    man_lvl_finish_player_not_found:
        call man_game_mark_as_beaten
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 6 Manager - Save checkpoint                                 ;;
;;  Save how far the player has made it into the level by saving    ;;
;;  the memory address of the last level it completed               ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX and BC                                      ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_save_checkpoint:
    ;; Save current level event at the checkpoint, to be restored later if the player dies
    call man_lvl_get_evt_size       ;; BC = current event size
    add ix, bc                      ;; Move IX to event next to checkpoint
    ld (lvl_reader_checkpoint), ix  ;; Save checkpoint there

    ;; Save points in backup
    call sys_score_save_score_state
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 7 Manager - Stop Level Reading                              ;;
;;  Mark the level reader to stop reading and wait for a game event ;;
;;  to tell us to keep reading                                      ;;
;; Input: -                                                         ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_stop_reading:
    ;; Init "keep reading" flag to read
    ld a, #0
    ld (lvl_reader_keep_reading), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Type 8 Manager - Show or Hide Boss string                        ;;
;;  Change the boss string to the specified sprite, and show it or  ;;
;;  hide it                                                         ;;
;; Input: -                                                         ;;
;; Output:  Destroys IX, A, BC, DE and HL                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
man_lvl_toggle_boss_title:
    ;; Check if a new title must be added, or the current title must be destroyed
    ld b, #hud_type_bossbar     ;; Load title filter
    push ix                         ;; Save IX backup
    call man_hud_search_type    ;; Search for titles
    pop ix                          ;; Load IX backup
    jr z, spawn_new_bosstitle       ;; If zero, no title found. A new title must be created

        ;; Delete title
        ld b, #hud_type_bossbar
        call man_mark_for_destroy_hud_of_type
        jp boss_title_spawn_skip

    spawn_new_bosstitle:
        ;; Select title template
        ld a, evt_param1(ix) ;; Param = Title to show (With sprite, anim, width height, xpos...)

        ;; Check if parameter is outside array size
        ld b, #bosstitle_selec_array_size;; B = array size
        inc b                           ;; B = array size + 1
        cp b                            ;; A - B+1 < 0 if A<=B (A is in range)
        jr c, bosstitle_param_fix_skip
            ld a, #1
        bosstitle_param_fix_skip:

        ;; Select bosstitle template (Load in HL)
        ld hl, #lvl_reader_bosstitle_select
        dec hl
        dec hl
        bosstitle_selection_loop:
            inc hl
            inc hl
            dec a
        jr nz, bosstitle_selection_loop

        ;; Load (HL) in HL
        ld c, (hl)
        inc hl
        ld b, (hl)
        ld h, b     ;; HL = Title template memory
        ld l, c

        ;; Create the title
        call man_create_one_hud_item
    boss_title_spawn_skip:
ret
