;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "ai.h.s"
.include "../manager/entities.h.s"
.include "../manager/bullet.h.s"
.include "render.h.s"
.include "death_behaviours.h.s"
.include "../animations/animations.h.s"

.globl cpct_getRandom_xsp40_u8_asm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Bullet - AI Behaviour Routine                                    ;;
;;  Bullet dies if it's touching the top or bottom of the screen    ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, B and HL                                     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_bullet:
   
    ld a, ent_y_pos(ix)         ;; Load position of the bullet
    cp #screen_y_start
    jr z, bullet_dies           ;; If bullet is at the top , it dies

    call sys_ai_check_bottom_of_screen
    jr nz, bullet_lives
     
    bullet_dies:

        ;; Change death behaviour to "Do Nothing" (to avoid death explosion)
        ld hl, #sys_death_donothing
        ld ent_death_beh(ix), l
        ld 1+ent_death_beh(ix), h

        call man_mark_one_entity_for_destruction

    bullet_lives:

ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Star - AI Behaviour Routine                                      ;;
;;  Check if it's at the bottom of the screen to go back up         ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys AF, B, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_star:
    call sys_render_mark_special_erasing
    call sys_ai_check_bottom_of_screen
    jr nz, star_keeps_going

        ;; Reset Y position
        ld a, #0
        ld ent_y_pos(ix), a

        ;; Get a random X [0-31]
        call cpct_getRandom_xsp40_u8_asm
        ld a, l
        ld b, #31
        and b
        ld ent_x_pos(ix), a

        ;; Get a random X [0-7]
        call cpct_getRandom_xsp40_u8_asm
        ld a, l
        ld b, #7
        and b

        ;; Add both Xs (31+7 -> [0-38])
        ld b, ent_x_pos(ix)
        add b

        ;; Increase by 1 -> [1-39]
        inc a
        ld ent_x_pos(ix), a

    star_keeps_going:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy - AI Behaviour Routine                                     ;;
;;  Enemy goes down until it disappears                             ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, B and HL                                     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_enemy:

    call sys_ai_check_bottom_of_screen
    jr nz, enemy_lives

        ;; Change death behaviour to "Do Nothing" (to avoid death explosion)
        ld hl, #sys_death_donothing
        ld ent_death_beh(ix), l
        ld 1+ent_death_beh(ix), h

        ;; We die (our IX is the input)
        call man_mark_one_entity_for_destruction

    enemy_lives:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Zigzag - AI Behaviour Routine                              ;;
;;  Enemy goes down until it disappears                             ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, B and HL                                     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_enemy_zigzagger:
    call sys_ai_change_x_vel_on_border
    call sys_ai_behaviour_enemy
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Shoots - AI Behaviour Routine                              ;;
;;  Extension to normal enemy behaviour, if it shoots               ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, BC and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_shoot:
    call sys_ai_update_shooting
    call sys_ai_behaviour_enemy
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Patroller - AI Behaviour Routine                           ;;
;;  Enemy that stays up shooting and moving horizontally            ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, BC and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_patrol:
    ;; Check if Y speed is zero (if it is, jump to patrol state)
    ld a, ent_y_vel(ix)
    or a
    jr z, patrol_phase

    ;; If not patrolling, check if patrol must start (that's when a certain height has been reached)
    ld a, ent_y_pos(ix)
    ld b, ent_ai_counter2(ix)
    cp b
    ret c ;; If Not Carry, then Ypos > 180, so it must keep going down

    ;; Start patrol: Reset Y speed, and initialize X speed
    ld a, #0
    ld ent_y_vel(ix), a
    ld a, ent_ai_counter(ix) ;; Init speed with what counter 1 says
    ld ent_x_vel(ix), a
    ;; Reuse counter 2 for X speed cap, and init counter 1 for shooting
    ld ent_ai_counter2(ix), a
    ld a, #enemy_shooting_cooldown
    ld ent_ai_counter(ix), a

    ;; Patrol (Move in X and shoot)
    patrol_phase:
        ;; Switch between (0 and 1)OR (0 and -1)
        ld a, ent_x_vel(ix)
        cp #0
        jr nz, speed_to_zero
            ;; If A is zero, set velocity to 1 or -1
            ld a, ent_ai_counter2(ix)
            ld ent_x_vel(ix), a
            jr speed_switch_end
        speed_to_zero:
            ;; If A is NOT zero, set velocity to zero
            ld a, ent_x_vel(ix)
            ld ent_ai_counter2(ix), a
            ld a, #0
            ld ent_x_vel(ix), a
        speed_switch_end:

        ;; Check shooting and X speed changes
        call sys_ai_update_shooting
        call sys_ai_change_x_vel_on_border
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Crawler - AI Behaviour Routine                             ;;
;;  Enemy that tours the screen (moves on Y, then X, then Y...)     ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, BC and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_crawl:
    ;; Check if we die
    call sys_ai_behaviour_enemy

    ;; Check if Y speed is zero (if it is, jump to patrol state)
    ld a, ent_y_vel(ix)
    or a
    jr nz, progress_phase

        ;; X phase
        ;;;;;;;;;;;

        ;; Switch between (0 and 1)OR (0 and -1)
        ld a, ent_x_vel(ix)
        cp #0
        jr nz, speed_to_zero_crawl
            ;; If A is zero, set velocity to 1 or -1
            ld a, ent_ai_counter2(ix)
            ld ent_x_vel(ix), a
            jr speed_switch_end_crawl
        speed_to_zero_crawl:
            ;; If A is NOT zero, set velocity to zero
            ld a, ent_x_vel(ix)
            ld ent_ai_counter2(ix), a
            ld a, #0
            ld ent_x_vel(ix), a
        speed_switch_end_crawl:

        ;; Check X speed changes because of borders
        call sys_ai_change_x_vel_on_border

        ;; If XVel != Counter2 (Speed != Previous Speed) then pass to phase Y
        ld a, ent_x_vel(ix)         ;; Speed        ;; |
        ld b, ent_ai_counter2(ix)   ;; PrevSpeed    ;; | If PrevSp==Sp, then 1+1=2 o -1-1=-2
        add b                       ;; 1 + (-1) = 0 ;; | If PrevSp!=Sp, then 1-1=0
        ret nz  ;; If non zero, keep on phase X

            ;; If it's zero, go to Y phase
            ld a, #0                            ;; X speed to zero
            ld ent_x_vel(ix), a
            ld a, #1                            ;; Y speed to one
            ld ent_y_vel(ix), a
            ld a, #enemy_crawling_y_cycles      ;; Counter to cycles that we'll go down
            ld ent_ai_counter2(ix), a

            ;; Shoot!
            call man_bullet_shoot_bullet

    ;; Patrol (Move in Y until counter reaches zero
    progress_phase:
        
        ;; Y phase
        ;;;;;;;;;;;

        ;; If not patrolling, check if patrol must start (that's when a certain height has been reached)
        ld a, ent_ai_counter2(ix)
        dec a
        ld ent_ai_counter2(ix), a
        ret nz ;; If not zero, keep going

            ;; If counter is zero, go to X phase
            ld a, #0                            ;; Y speed to zero
            ld ent_y_vel(ix), a
            ld a, #1                            ;; X speed default to 1
            ld ent_x_vel(ix), a
            call sys_ai_change_x_vel_on_border  ;; Sets X speed to 1 or -1, depending on border

            ;; Save X speed in X counter
            ld a, ent_x_vel(ix)
            ld ent_ai_counter2(ix), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Stalker - AI Behaviour Routine                             ;;
;;  Enemy that goes down trying to catch the player in X (slowly)   ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys IY, A, BC and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_stalker:
    ;; Check border reached
    call sys_ai_behaviour_enemy

    ;; Speed to zero
    ld a, #0
    ld ent_x_vel(ix), a

    ;; Check if we move on X
    ld a, ent_ai_counter2(ix)   ;; Invert A from 0 to 1 or viceversa
    inc a
    and #3 ;; 0b00000011
    ld ent_ai_counter2(ix), a
    ret nz                       ;; Do nothing when is zero

    ;; Move on X

        ;; Search the player (Player data to IX)
        push ix     ;; Save enemy's data on stack
        ld b, #ent_type_player
        call man_search_type
        push ix
        pop iy      ;; IY = Player's data
        pop ix      ;; IX = Enemy's data
        ret z       ;; If player was not found, do nothing

        ;; Check if we're left or right from them, or centered
        ld a, ent_x_pos(ix) ;; A = Enemy's  X pos
        ld b, ent_x_pos(iy) ;; B = Player's X pos
        sub b

        ;; Assign needed X speed
            ;; A > B -> NC  Player is to my left
            ;; A = B -> Z
            ;; A < B -> C   Player is to my right
        ret z   ;; If Z, A=B, so we don't need to move
        jp nc, player_is_left
        jp c,  player_is_right
ret
player_is_left:
    ld a, #-1
    ld ent_x_vel(ix), a
    ret
player_is_right:
    ld a, #1
    ld ent_x_vel(ix), a
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Enemy Flag - AI Behaviour Routine                                ;;
;;  Goes down and is "picked up" by the player and kills all enemies;;
;;  If it reaches the bottom, kills the player                      ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, B and HL                                     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_grabflag:
    ;; Switch Y speed between 0 and 1
    ld a, ent_y_vel(ix)
    xor #1
    ld ent_y_vel(ix), a

    ;; Check position to change animation
    ld a, ent_y_pos(ix)
    cp #powercell_change1_y
    jr nz, skip_fast_anim
        ;; Set fast animation
        ld hl, #anim_enemy_flag_fast
        ld ent_anim(ix),   l
        ld 1+ent_anim(ix), h
    skip_fast_anim:
    cp #powercell_change2_y
    jr nz, skip_faster_anim
        ;; Set faster animation
        ld hl, #anim_enemy_flag_faster
        ld ent_anim(ix),   l
        ld 1+ent_anim(ix), h
    skip_faster_anim:

    ;; Check if we reached the bottom
    call sys_ai_check_bottom_of_screen
    ret nz  ;; If not at the bottom, return

        ;; Change death behaviour to "Flag Escaped" (to kill the player and not the enemies)
        ld hl, #sys_death_boss_flag_escaped
        ld ent_death_beh(ix), l
        ld 1+ent_death_beh(ix), h

        ;; We die (our IX is the input)
        call man_mark_one_entity_for_destruction

ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Autodestroy - AI Behaviour Routine                               ;;
;;  Decrease counter until zero, then dies                          ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A and HL                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_autodestroy:
    ;; Decrease AI Counter
    ;; If AI counter is zero, we die
    ld a, ent_ai_counter(ix)
    dec a
    jr nz, ai_autodestroy_deathskip
        ;; Change death behaviour to "Do Nothing" (to avoid death explosion)
        ld hl, #sys_death_donothing
        ld ent_death_beh(ix), l
        ld 1+ent_death_beh(ix), h
        ;; We die (our IX is the input)
        call man_mark_one_entity_for_destruction

    ;; Else, save counter and move on
    ai_autodestroy_deathskip:
    ld ent_ai_counter(ix), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Player Escape - AI Behaviour Routine                             ;;
;;  AI-controlled player that goes out of the screen                ;;
;; Input: IX with entity's memory position                          ;;
;; Output: ???                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_behaviour_player_escape:
    ;; Check phase
    ld a, ent_ai_counter(ix)
    and #0x01
    jr z, check_escape_phase_2
        ;; Check if speed must be negative or positive to reach the center
        ld a, ent_x_vel(ix)
        or a
        jp nz, set_speed_to_zero
            ld a, ent_x_pos(ix)
            cp #16
            jr nc, set_x_speed_neg
                ;; We are to the left, positive speed
                ld a, #1
                ld ent_x_vel(ix), a
                jp move_now
            set_x_speed_neg:
                ;; We are to the right, negative speed
                ld a, #-1
                ld ent_x_vel(ix), a
            move_now:
            jp x_speed_setted
        set_speed_to_zero:
            ld a, #0
            ld ent_x_vel(ix), a
        x_speed_setted:

        ;; Go down a bit
        ld a, #1
        ld ent_y_vel(ix), a

        ;; Check if we are at the screen center
        ld a, ent_x_pos(ix)
        cp #16 ;; Screen center for player
        ret nz  ;; If non zero (X != 16) do nothing

            ;; Init next phase
            ld a, #0
            ld ent_x_vel(ix), a
            ld a, #-1
            ld ent_y_vel(ix), a
            ld a, #70
            ld ent_ai_counter2(ix), a
            ;; Change phase
            ld a, #0x02
            ld ent_ai_counter(ix), a

        ret
    check_escape_phase_2:
    ld a, ent_ai_counter(ix)
    and #0x02
    jr z, check_escape_phase_3
        ;; Tilt Y speed
        ld a, ent_y_vel(ix)
        neg
        ld ent_y_vel(ix), a

        ;; Check phase change
        ld a, ent_ai_counter2(ix)
        dec a
        ld ent_ai_counter2(ix), a
        ret nz
            ;; Change phase
            ld a, #0x04
            ld ent_ai_counter(ix), a

        ret
    check_escape_phase_3:
    ld a, ent_ai_counter(ix)
    and #0x04
    ret z
        ;; Increase speed (Decrease to go up)
        ld a, ent_y_vel(ix)
        dec a

        ;; Cap at 7
        neg
        and #7 ;; 0b0000 0111
        neg
        ld ent_y_vel(ix), a

        ;; Check if we reached the top of the screen
        ld a, ent_y_pos(ix) ;; Load position of the player-AI
        cp #screen_y_start
        ret nz              ;; If we're at the top, finish the game (disappear/die)
            ;; We're at the top
            call man_mark_one_entity_for_destruction ;; Autodestroy
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Change X - Auxiliar Routine                                      ;;
;;  Changes X vel to 1 if we're on left border, to -1 if left border;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A and B                                         ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_change_x_vel_on_border:
    ld a, #screen_width         ;; Load screen height
    ld b, ent_spr_w(ix)         ;; Load enemy height

    sub b                       ;; We get the max posible Y

    ld b, ent_x_pos(ix)         ;; Load enemy position

    sub b                       ;; If screen height - enemy position = 0, enemy is at the bottom of the screen

    jr nz, enemy_end_stop_r

        ld a, #0xFF
        ld ent_x_vel(ix), a

    enemy_end_stop_r:
    
    ;;has the X pos
    inc b
    dec b

    jr nz, enemy_end_stop_l

        ld a, #0x01
        ld ent_x_vel(ix), a

    enemy_end_stop_l:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check Bottom - Auxiliar Routine                                  ;;
;;  Checks if an entity is at the bottom of the screen              ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Z if entity's at bottom, NZ if not                       ;;
;;         Destroys A and B                                         ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_check_bottom_of_screen:
    ld a, #screen_height        ;; Load screen height
    ld b, ent_spr_h(ix)         ;; Load bullet height
    sub b                       ;; We get the max posible Y

    ld b, ent_y_pos(ix)         ;; Load bullet position
    sub b                       ;; If screen height - bullet position = 0, bullet is at the bottom of the screen
    ;; Z  if true  (we're at bottom)
    ;; NZ if false (not   at bottom)
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Shooting - Auxiliar Routine                               ;;
;;  Uses counter 1 to update shooting cooldown, then uses the       ;;
;;  entity's attack patter to shoot                                 ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A, BC and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_update_shooting:
    ld a, ent_ai_counter(ix)
    dec a
    ld ent_ai_counter(ix), a
    jp nz, shooted                          ;; If cooldown is 0, it shoots

        ld a, #enemy_shooting_cooldown      ;; Reset cooldown

        ld ent_ai_counter(ix), a            ;;

        call man_bullet_shoot_bullet  ;; Create a bullet    


    shooted:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update AI of One Entity                                          ;;
;;  Updates the ai for the specified entity, using its AI Behaviour ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys HL and BC, Might Destroy everything else        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_update_one_entity:
    ;; Load AI Behaviour routine's memory position into HL
    ld l, ent_ai_behave(ix)
    ld h, 1+ent_ai_behave(ix)

    ;; Load return point
    ld bc, #ai_update_one_return_point
    push bc

    ;; Call Behaviour routine
    jp (hl)

    ;; Return point
    ai_update_one_return_point:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update AI of all Entity                                          ;;
;;  Asks the manager to update the AI of every AI entity            ;;
;; Input:                                                           ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_ai_update_all_entities:
    ld hl, #sys_ai_update_one_entity ;; Load routine's memory position
    ld b, #ent_cmp_ai                ;; Load filter mask (only AI entities)
    call man_for_all_matching
ret
