;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "collisions.h.s"
.include "../manager/entities.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Collisions of a Pair of Entities                          ;;
;;  Checks collisions of two given entities. checks if they can     ;;
;;  collide and does stuff                                          ;;
;; Input: IX and IY with entity's memory position                   ;;
;; Output: -                                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_collisions_update_two_entities:
    ;; Init Collision Register
    ld de, #0x0000

    ;; Checks if IX collides with IY's type/layer
    ld a, ent_coll_mask(ix) ;; Load Collision mask
    ld b, ent_type(iy)      ;; Load type
    and b                   ;; AND Checks if at least one type collides
    jr z, skip_ix_with_iy   ;; If a collision is possible, check consequences

        ;; IX can collide with IY
        ld d, #1    ;; Mark IX as collidable

    skip_ix_with_iy:

    ;; Checks if IY collides with IX's type/layer
    ld a, ent_coll_mask(iy) ;; Load Collision mask
    ld b, ent_type(ix)      ;; Load type
    and b                   ;; AND Checks if at least one type collides
    jr z, skip_iy_with_ix   ;; If a collision is possible, check consequences

        ;; IY can collide with IX
        ld e, #1    ;; Mark IY as collidable

    skip_iy_with_ix:

    ;; Check if IX collides physically with IY (DE != 0)
    ld a, #0x00 ;; A = all bits to 1
    add d       ;; D = 0 if IX does not collide with IY, else is !=0
    add e       ;; E = 0 if IX does not collide with IY, else is !=0
    jr z, no_interaction  ;; If FF AND E = 0 -> Then none collides with none

        ;; Interaction is possible! Check geometric calculations
        call sys_collisions_check_collision ;; NZ If no collision (Does not destroy DE)
        jr nz, no_physical_collision

            ;; If D!=0, IX collided with IY
            ld a, #0xFF ;; A = all bits to 1
            and d       ;; D = 0 if IX does not collide with IY, else is !=0
            jr z, iy_collision_with_ix ;; If AND = 0, no collision, skip to IY with IX

                ;; Decrease IX's health count
                dec ent_health(ix)    ;; Load health in A
                jr nz, iy_collision_with_ix ;; If health is zero, kill the entity
                    call man_mark_one_entity_for_destruction ;; IX is the input

            ;; If E!=0, IY collided with IX
            iy_collision_with_ix:
            ld a, #0xFF ;; A = all bits to 1
            and e       ;; E = 0 if IX does not collide with IY, else is !=0
            jr z, no_physical_collision ;; If AND = 0, no collision

                ;; Decrease IY's health count
                dec ent_health(iy)    ;; Load health in A
                jr nz, no_physical_collision ;; If health is zero, kill the entity
                    push ix     ;; Save backup of IX
                    push iy     ;; Store IY to pass it to IX
                    pop ix      ;; Copy IY to IX
                    call man_mark_one_entity_for_destruction ;; IX is the input
                    pop ix      ;; Recover IX backup

        no_physical_collision:
    no_interaction:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Collisions of All Entities                                ;;
;;  Asks the manager to check collisions between                    ;;
;;  every pair of entities                                          ;;
;; Input: IX and IY with entity's memory position                   ;;
;; Output: Destroys IX, IY, A, B, DE and HL                         ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_collisions_update_all_entities:
    ld hl, #sys_collisions_update_two_entities  ;; Load routine's memory position
    ld b, #ent_cmp_collider                     ;; Load filter mask (only physics entities)
    call man_for_all_pairs_matching             
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Check collision between to collidable entities                   ;;
;;  Checks if two given entities collide                            ;;
;; Input: IX and IY with entity's memory position                   ;;
;; Output: Z if collision, NZ if not.    Destroys A, B and HL       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_collisions_check_collision:
    ;; Save DE
    push de

    ;; Init positions in cheaper registers
    ld h, ent_x_pos(iy)
    ld l, ent_x_pos(ix)
    ld d, ent_spr_w(iy)
    ld e, ent_spr_w(ix)

    call sys_collisions_adjust_player_x_ix

    ;; Check E1.x < E2.x + E2.width
    ld a, h             ;; E2.x
    ld b, d             ;; E2.width
    dec b               ;; Adjust to width (x + width - 1)
    add b               ;; A = E2 -> x+w
    ld b, l             ;; B = E1 -> x
    sub b               ;; A - B = E2 - E1 >= 0 (nc)
    jr c, no_collision  ;; If Carry, E2 < E1, then is false

    ;; Check E1.x + E1.width > E2.x
    ld a, l             ;; E1.x
    ld b, e             ;; E1.width
    dec b               ;; Adjust to width (x + width - 1)
    add b               ;; A = E1 -> x+w
    ld b, h             ;; B = E2 -> x
    sub b               ;; A - B = E1 - E2 >= 0 (nc)
    jr c, no_collision  ;; If Carry, E1 < E2, then is false

    ;; Init positions in cheaper registers
    ld h, ent_y_pos(iy)
    ld l, ent_y_pos(ix)
    ld d, ent_spr_h(iy)
    ld e, ent_spr_h(ix)

    call sys_collisions_adjust_player_y_ix

    ;; Check E1.y < E2.y + E2.height
    ld a, h             ;; E2.y
    ld b, d             ;; E2.height
    dec b               ;; Adjust to height (y + height - 1)
    add b               ;; A = E2 -> y+h
    ld b, l             ;; B = E1 -> y
    sub b               ;; A - B = E2 - E1 >= 0 (nc)
    jr c, no_collision  ;; If Carry, E2 < E1, then is false

    ;; Check E1.y + E1.height > E2.y
    ld a, l             ;; E1.y
    ld b, e             ;; E1.height
    dec b               ;; Adjust to height (y + height - 1)
    add b               ;; A = E1 -> y+h
    ld b, h             ;; B = E2 -> y
    sub b               ;; A - B = E1 - E2 >= 0 (nc)
    jr c, no_collision  ;; If Carry, E1 < E2, then is false

    ;; If there is collision, A = 0 so Z = 1
    ld a, #0x00
    jr collision_check_end

    ;; If there is not a collision, A = 1 so Z = 0
    no_collision:
    ld a, #0x01

    collision_check_end:
    pop de  ;; Recover DE
    or a
    ;; Z if collision, NZ if not collision
ret

sys_collisions_adjust_player_x_ix:
    ;; Check player type
    ld a, ent_type(ix)
    and #ent_type_player
    ret z

        ;; Increase Xpos (L) by 1
        inc l
        ;; Decrease Width (E) by 2
        dec e
        dec e
ret

sys_collisions_adjust_player_y_ix:
    ;; Check player type
    ld a, ent_type(ix)
    and #ent_type_player
    ret z

        ;; Increase Ypos (L) by 1
        inc l
        ;; Decrease Height (E) by 2
        dec e
        dec e
ret

; sys_collisions_adjust_player_iy:
;     ;; Check player type
;     ld a, ent_type(iy)
;     and #ent_type_player
;     ret z

;         ;; Increase Xpos (L) by 1
;         inc h
;         ;; Decrease Width (E) by 2
;         dec d
;         dec d
; ret
