;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "cpct_override.h.s"
.include "cpctelera.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Get Screen Pointer (20x25 char screen)                           ;;
;;  Returns a video memory position, given an (X, Y) screen position;;
;; Input:   DE with video memory's starting position                ;;
;;          C  with X position                                      ;;
;;          B  with Y position                                      ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
cpcto_getScreenPtr_20x25:
    ;; Add up X coordinate to screen_memory, so that DE = (screen_start + x)
    ;; (DE += C, as DE = screen_memory, and C = x coordinate) 
    ld    a, e     ;; [1] / E = E + x   (As C = x)
    add   c        ;; [1] |   ( If E + x > 0xFF, Carry is set. We must Add it to D )
    ld    e, a     ;; [1] |
    adc   d        ;; [1] | D = D + (C + A) + Carry   (E = C + A)
    sub   e        ;; [1] | D = D + Carry
    ld    d, a     ;; [1] |   (If E + x did set carry, this will do D += 1, else D += 0)

    ;; Let y' = [ y / 8 ] = int(y / 8) (Integer division by 8)
    ;; Calculate 40y' = (32 + 8)y' (0x28 * y')
    ld    a, b     ;; [1] A = y coordinate (which is stored in b)
    and   #0xF8    ;; [2] A = 8y'   ( and 0xF8 is the same as 8 * int(y / 8) = 8y' )
    ld    l, a     ;; [1] | HL = 8y'  (H = 0, L = 8y')
    ld    h, #0    ;; [2] |
    ld    a, b     ;; [1] A = y coordinate (Save for later use, freeing BC)
    ld    b, h     ;; [1] | BC = 8y'
    ld    c, l     ;; [1] |
    add  hl, hl    ;; [3] HL = 8y' + 8y'   = 16y'
    add  hl, hl    ;; [3] HL = 16y' + 16y' = 32y'
    add  hl, bc    ;; [3] HL = 32y' + 8y'  = 40y'

    ;; Calculate 2048 * (y % 8)  (0x800 * (y % 8))
    and   #0x07    ;; [2]  A = y % 8  (A had y coordinate)
    rlca           ;; [1]  A = 2 (y % 8)
    rlca           ;; [1]  A = 4 (y % 8)
    rlca           ;; [1]  A = 8 (y % 8)
    ;; Now [A 0x00] (Using a as HSB) would be 0x100 * A = 0x100 * 8 (y % 8 )
    ;; So, we only have to add A to H in order to have HL += 0x100 * 8 (y % 8)
    add   h        ;; [1]  | HL += 0x100 * 8 (y % 8)
    ld    h, a     ;; [1]  |

    ;; Add up everything and return
    add  hl, de    ;; [3] HL = 40 * [y / 8] + 2048 * (y % 8) + (screen_start + x)
ret 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Draw Sprite (20x25 char screen)                                  ;;
;;  Draws a sprite given its data, position, width and height       ;;
;;  for a 20x25 character screen                                    ;;
;; Input:   DE with upper left corner's video memory pointer        ;;
;;          HL with sprite's array starting memory position         ;;
;;          C  with sprite's width                                  ;;
;;          B  with sprite's height                                 ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
cpcto_drawSprite_20x25:
    ;; Modify code using width to jump in drawSpriteWidth
    ld    a, #126           ;; [2] We need to jump 126 bytes (63 LDIs*2 bytes) minus the width of the sprite * 2 (2B)
    sub   c                 ;; [1]    to do as much LDIs as bytes the Sprite is wide
    sub   c                 ;; [1]
    ld (ds_drawSpriteWidth+#4), a ;; [4] Modify JR data to create the jump we need

    ld    a, b              ;; [1] A = Height (used as counter for the number of lines we have to copy)
    ex   de, hl             ;; [1] Instead of jumping over the next line, we do the inverse operation because 
                            ;; .... it is only 4 cycles and not 10, as a JP would be)

    ds_drawSpriteWidth_next:
    ;; NEXT LINE
    ex   de, hl             ;; [1] HL and DE are exchanged every line to do 16bit maths with DE. 
                            ;; .... This line reverses it before proceeding to copy the next line.
    ds_drawSpriteWidth:
    ;; Draw a sprite-line of n bytes
    ld   bc, #0x800  ;; [3] 0x800 bytes is the distance in memory from one pixel line to the next within every 8 pixel lines
                     ;; ... Each LDI performed will decrease this by 1, as we progress through memory copying the present line
    jr__0            ;; [3] Self modifying instruction: the '00' will be substituted by the required jump forward. 
                     ;; ... (Note: Writting JR 0 compiles but later it gives odd linking errors)
    ldi              ;; [5] <| 63 LDIs, which are able to copy up to 63 bytes each time.
    ldi              ;; [5]  | That means that each Sprite line should be 63 bytes width at most.
    ldi              ;; [5]  | The JR instruction at the start makes us ignore the LDIs we don't need 
    ldi              ;; [5] <| (jumping over them) That ensures we will be doing only as much LDIs 
    ldi              ;; [5] <| as bytes our sprite is wide.
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    ldi              ;; [5] <|
    ldi              ;; [5] <|
    ldi              ;; [5]  |
    ldi              ;; [5]  |
    
    dec   a          ;; [1] Another line finished: we discount it from A
    ret   z          ;; [2/4] If that was the last line, we safely return

    ;; Jump destination pointer to the start of the next line in video memory
    ex   de, hl      ;; [1] DE has destination, but we have to exchange it with HL to be able to do 16bit maths
    add  hl, bc      ;; [3] We add 0x800 minus the width of the sprite (BC) to destination pointer 
    ld    b, a       ;; [1] Save A into B (B = A)
    ld    a, h       ;; [1] We check if we have crossed video memory boundaries (which will happen every 8 lines). 
                     ;; .... If that happens, bits 13,12 and 11 of destination pointer will be 0
    and   #0x38      ;; [2] leave out only bits 13,12 and 11 from new memory address (00xxx000 00000000)
    ld    a, b       ;; [1] Restore A from B (A = B)
    jp   nz, ds_drawSpriteWidth_next ;; [3] If any bit from {13,12,11} is not 0, we are still inside 
                                     ;; .... video memory boundaries, so proceed with next line

    ;; Every 8 lines, we cross the 16K video memory boundaries and have to
    ;; reposition destination pointer. That means our next line is 16K-0x28 bytes back
    ;; which is the same as advancing 48K+0x28 = 0xC028 bytes, as memory is 64K 
    ;; and our 16bit pointers cycle over it
    ld   bc, #0xC028           ;; [3] We advance destination pointer to next line
    add  hl, bc                ;; [3]  HL += 0xC028
    jp ds_drawSpriteWidth_next ;; [3] Continue copying
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Draw Solid Box (20x25 char screen)                               ;;
;;  Draws a solid box given its position, width, height and         ;;
;;  color pattern in a 20x25 character screen                       ;;
;; Input:   DE with upper left corner's video memory pointer        ;;
;;          C  with box's width                                     ;;
;;          B  with box's height                                    ;;
;;          A  with box's color pattern, like painting a single byte;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
cpcto_drawSolidBox_20x25:
    ;; COPY2HL_n_INC Macro 
    ;;  Copy Register To (HL) and Increment HL
    ;; INPUT:
    ;;    reg8 - any 8-bits register
    ;;
    .macro COPY2HL_n_INC reg8
        ld   (hl), reg8         ;; [2] (HL) = reg8
        inc   hl                ;; [2] HL++  
    .endm

    ;; DrawSolidBox code starts here (immediately after getting parameters from stack)
    ld     h, d             ;; [1] / HL = DE  
    ld     l, e             ;; [1] \ HL Points to Video Memory, and DE holds a copy which won't be modified
    ld     e, a             ;; [1] E=A (Colour pattern)
    
    ;; Modify code using width to jump to ensure
    ;; that only width bytes are copied each line 
    ld     a, #128          ;; [2] We need to jump 128 bytes (64 COPY2HL_n_INC*2 bytes) minus the width of the sprite * 2 (2B)
    sub    c                ;; [1]  to do as much COPY2HL_n_INC as bytes the Sprite is wide
    sub    c                ;; [1]
    ld (_jr_offset), a      ;; [4] Modify JR data to create the jump we need
    
    ld     c, e             ;; [1] C=E (Colour pattern)
    ld     e, l             ;; [1] Restore DE = HL

    _next_line:
    _jr_offset = .+1
   jr__0                   ;; [3] Self modifying instruction: the '00' will be substituted by the required jump forward. 
                           ;; ... (Note: Writting JR 0 compiles but later it gives odd linking errors)  

    ;; 64 COPY2HL_n_INC c, which are able to copy up to 64 bytes each time.
    ;; That means that each Sprite line should be 64 bytes width at most.
    ;; The JR instruction at the start makes us ignore the COPY2HL_n_INC we don't need 
    ;; (jumping over them) That ensures we will be doing only as much COPY2HL_n_INC 
    ;; as bytes our sprite is wide.
    .rept 64
        COPY2HL_n_INC c         ;; [4] (HL) = c, ++HL { Repeated 64 times }
    .endm
    
    dec    b                ;; [1] Another line finished: we discount it from D
    ret    z                ;; [2/4] If that was the last line, we safely return

    ;; Jump destination pointer to the start of the next line in video memory
    ld     a, #0x8          ;; [2] / HL = DE = DE + 0x800
    add    d                ;; [1] | DE holds the pointer to the start of this line
    ld     h, a             ;; [1] | Adding 0x800 makes HL point to the start of
    ld     d, a             ;; [1] | the next line to be filled in with values
    ld     l, e             ;; [1] \ DE holds a copy which won't be modified
    ;; We check if we have crossed video memory boundaries (which will happen every 8 lines). 
    ;; ... If that happens, bits 13,12 and 11 of destination pointer will be 0
    and   #0x38             ;; [2] leave out only bits 13,12 and 11 from new memory address (00xxx000 00000000)
    jp    nz, _next_line    ;; [3] If any bit from {13,12,11} is not 0, we are still inside 
                            ;; ... video memory boundaries, so proceed with next line

    ;; Every 8 lines, we cross the 16K video memory boundaries and have to
    ;; reposition destination pointer. That means our next line is 16K-0x50 bytes back
    ;; which is the same as advancing 48K+0x50 = 0xC050 bytes, as memory is 64K 
    ;; and our 16bit pointers cycle over it
    ld    hl, #0xC028       ;; [3] We advance destination pointer to next line
    add   hl, de            ;; [3] HL = DE + 0xC050
    ld     d, h             ;; [1] / DE = HL
    ld     e, l             ;; [1] \
    jp   _next_line         ;; [3] Continue copying   
ret
