;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "death_behaviours.h.s"
.include "../manager/entities.h.s"
.include "../manager/game.h.s"
.include "../manager/level_reader.h.s"
.include "../system/score.h.s"
.include "../system/audio.h.s"
.include "../templates/templates.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; WARNING! NONE OF THIS ROUTINES SHALL DESTROY IX!!!!!!   ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Null - Death Behaviour Routine                                   ;;
;;  Does literally nothing                                          ;;
;; Input: IX with entity's memory position                          ;;
;; Output: -                                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_donothing:
    ;; Does nothing
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Boss - Death Behaviour Routine                                   ;;
;;  All bosses do this to enable level reading to continue.         ;;
;;  This is beacuse bosses stop level reading to avoid time attack  ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_boss:
    ;;call man_lvl_continue_reading
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Boss Core - Death Behaviour Routine                              ;;
;;  Defended cores kill al of their units upon death                ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_boss_core:
    ;; Create large explosion
    call sys_death_mid_explode

    ;; Destroy all enemies
    call man_game_activate_enemy_extinction

    ;; Generic Boss behaviour
    call sys_death_boss
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Boss Flag - Death Behaviour Routine                              ;;
;;  "Grab the flag" bosses lose if you grab the flag, by win if     ;;
;;  they get to the end. Flag entity changes its death behaviour    ;;
;;  automatically if they die or reach the end                      ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_boss_flag_pickedup:
    ;; Same as boss core
    jp sys_death_boss_core
ret
sys_death_boss_flag_escaped:
    ;; Destroy the player
    call man_game_activate_player_extinction
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Explosion - Death Behaviour Routine                       ;;
;;  Creates a mid explosion on its position                         ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys IY, A and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_small_explode:
    ;; Create its death explosion
    ld hl, #ent_template_small_explosion
    call aux_create_explosion
    
    call sys_death_earn_points
ret
sys_death_mid_explode:
    ;; Create its death explosion
    ld hl, #ent_template_mid_explosion
    call aux_create_explosion

    ;; Throw sound
    call sys_audio_play_sfx_explosion

    call sys_death_earn_points
ret
sys_death_midmid_explode:
    ;; Create its death explosion
    ld hl, #ent_template_midmid_explosion
    call aux_create_explosion

    ;; Throw sound
    call sys_audio_play_sfx_explosion

    call sys_death_earn_points
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Player Explosion - Death Behaviour Routine                ;;
;;  Creates a player explosion on their position                    ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys IY, A and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_player_explode:
    ;; Create its death explosion
    ld hl, #ent_template_player_explosion
    call aux_create_explosion
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Player Escaped - Death Behaviour Routine                         ;;
;;  Called when player finishes the game                            ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A                                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_player_escaped:
    ;; Mark the game as beaten
    call man_game_mark_as_beaten
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ExtraLife Powerup - Death Behaviour Routine                      ;;
;;  Adds an extra life to the player                                ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_extralife:
    push ix
    call man_game_add_extralife
    pop ix
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Upgrade Powerup - Death Behaviour Routine                        ;;
;;  Changes player's attack pattern to next level                   ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A, B, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_wpn_upgrade:
    push ix
    call man_game_weapon_up
    pop ix
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Points Powerup - Death Behaviour Routine                         ;;
;;  Awards the player with points                                   ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys A and HL                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_earn_points:
    ;; Input = DEBC = 0000BB00
    ld de, #0x0000
    ld  b, ent_points(ix)
    ld  c, #0
    call sys_score_add
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Bomb Powerup - Death Behaviour Routine                           ;;
;;  Kills all enemies in the screen                                 ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_bomb:
    ;; Destroy all enemies
    call man_game_activate_enemy_extinction
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Player - Death Behaviour Routine                                 ;;
;;  Tell the game to start the death animation                      ;;
;; Input: IX with entity's memory position                          ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_death_player_death:
    call sys_death_player_explode
    call man_game_start_player_death_animation
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; AUXILIAR                    ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Explosion From Template                                   ;;
;;  Creates an explosion given an entity and a template             ;;
;; Input: IX with entity's memory position and HL with template     ;;
;; Output: IY and A                                                 ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
aux_create_explosion:
    ;; Create dead entity's backup
    push ix

    ;; Create its death explosion
    ;; HL stores the explosion template
    call man_create_one_entity

    ;; Modify its position
    pop iy              ;; Recover dead entity in IY
    ld a, ent_x_pos(iy) ;; |
    ld ent_x_pos(ix), a ;; | Copy dead's position to explosion position
    ld a, ent_y_pos(iy) ;; |
    ld ent_y_pos(ix), a ;; | 

    ;; Recover dead entity's backup
    push iy
    pop ix
ret
