;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "physics.h.s"
.include "../manager/game.h.s"
.include "../manager/entities.h.s"
.include "render.h.s"
.include "input.h.s"
.include "../templates/templates.h.s"
.include "cpctelera.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Physics of One Entity                                     ;;
;;  Updates the physics for the specified entity, including speed   ;;
;;  and input                                                       ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_physics_update_one_entity:
    ;;;; Input check
    ;;;;;;;;;;;;;;;;
        ;; Check if entity is controlled by input
        ld a, ent_comps(ix)             ;; Load entity's component mask
        and #ent_cmp_input              ;; Bitwise AND is Zero if input bit is not fulfilled
        jr z, physics_update_one_skip_input_check  ;; If Zero, no input must be checked, so we skip it

            ;; Reset entity's speed to zero (to avoid movement if no key is pressed)
            ld a, #0x00
            ld ent_x_vel(ix), a
            ld ent_y_vel(ix), a

            ;; Check and manage input
            call sys_input_check

    physics_update_one_skip_input_check:
    ;;;; Save previous position
    ;;;;;;;;;;;;;;;;;;;
    ld b, ent_x_pos(ix)
    ld c, ent_y_pos(ix)
    push bc

    ;;;; Move in X axis
    ;;;;;;;;;;;;;;;;;;;

        ;; Get X pos and X speed
        ld a, ent_x_pos(ix)
        ld b, ent_x_vel(ix)

        ;; Add X speed to X pos
        add b               ;; A = A + B (Xpos += Xspeed)
        ld ent_x_pos(ix), a ;; Save on memory

        ;; Check if new X position left the screen (out of 0-79)
        ld a, #256-screen_width ;; A = 256 - screen_w = size of out of screen
        ld b, ent_spr_w(ix)     ;; Load sprite width in B
        add b                   ;; Add that width to the margin size
        ld b, a                 ;; Save that on B
        ; Now B has the size of the out of screen (from (screen_width - sprite_width) to 255)

        ld a, ent_x_pos(ix) ;; Reload X pos in A
        add b               ;; Add that to A (if A+(256-80) > 255, then A>=0x80 and Carry flag is 1)
        jr nc, physics_update_one_skip_x_correction ;; If Carry flag is zero, it's fine
            call sys_render_mark_special_erasing

            ;; If X left the screen, correct the position (capped at 0 or 80-width)
            ;; Check if speed is positive or negative (speed > 127 or speed > 0x7F) (to cap at 0 or 80-w)
            ld a, ent_x_vel(ix) ;; Load speed in A
            or a
            jp m, correct_x_pos_negative  ;; Carry flag is 1 if speed > 127
                ;; Correction if positive speed
                ld a, #screen_width ;; Load right screen margin
                ld b, ent_spr_w(ix) ;; Load entity's width
                sub b               ;; Decrease margin by entitiy's width
                ld ent_x_pos(ix), a ;; Save capped position in memory
                jr physics_update_one_skip_x_correction

                ;; Correction if negative speed
                correct_x_pos_negative:
                ld a, #0x00         ;; Load left screen margin
                ld ent_x_pos(ix), a ;; Save capped position in memory

        physics_update_one_skip_x_correction:

    ;;;; Move in Y axis
    ;;;;;;;;;;;;;;;;;;;

        ;; Get Y pos and y speed
        ld a, ent_y_pos(ix)
        ld b, ent_y_vel(ix)

        ;; Add Y speed to Y pos
        add b               ;; A = A + B (Ypos += Yspeed)
        ld ent_y_pos(ix), a ;; Save on memory

        ;; Check if new Y position left the screen (out of 0-79)
        ld a, #256-screen_height;; Load screen height in B
        ld b, ent_spr_h(ix) ;; Load sprite height in B
        add b               ;; Add that height to the margin size
        ld b, a             ;; Save that on B
        ; Now B has the size of the out of screen (from (screen_height - sprite_height) to 255)

        ld a, ent_y_pos(ix) ;; Reload Y pos in A
        add b               ;; Add that to A (if A+(256-200) > 255, then A>=0xC8 and Carry flag is 1)
        jr nc, phys_update_skip_positive_y_correction ;; If Not Carry, it's fine
            ;; Mark for special erasing
            call sys_render_mark_special_erasing

            ;; Correction if positive speed
            ld a, #screen_height;; Load bottom screen margin
            ld b, ent_spr_h(ix) ;; Load entity's height
            sub b               ;; Decrease margin by entitiy's height
            ld ent_y_pos(ix), a ;; Save capped position in memory
            jr physics_update_one_skip_y_correction
        
        phys_update_skip_positive_y_correction:

        ;; Check if we're before the screen start
        ld a, ent_y_pos(ix)         ;; A = Ypos
        ld b, #screen_y_start       ;; B = Screen start point
        cp b                        ;; A - B
        jr nc, physics_update_one_skip_y_correction ;; If Ypos >= Start, there's not carry
            
            ;; Correction if negative speed
            correct_y_pos_negative:
            ld a, #screen_y_start   ;; Load upper screen margin
            ld ent_y_pos(ix), a     ;; Save capped position in memory

            call sys_render_mark_special_erasing

        physics_update_one_skip_y_correction:

    ;;;; Save previous position offset
    ;;;;;;;;;;;;;;;;;
    pop bc  ;; Get prev pos
    ld a, b
    ld b, ent_x_pos(ix)
    sub b               ;; A = previousX - X
    ld d, a             ;; Save A
    ld b, ent_prev_x(ix);; B = prevX
    add b               ;; A = prevX(now) + prevX(old) = prevprevX
    ld ent_prev_prev_x(ix), a
    ld ent_prev_x(ix), d
    
    ld a, c
    ld c, ent_y_pos(ix)
    sub c               ;; A = previousY - Y
    ld d, a             ;; Save A
    ld c, ent_prev_y(ix);; B = prevY
    add c               ;; A = prevY(now) + prevY(old) = prevprevY
    ld ent_prev_prev_y(ix), a
    ld ent_prev_y(ix), d
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Physics of All entites                                    ;;
;;  Updates the physics of every physic-dependant entity            ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_physics_update_all_entities:
    ld hl, #sys_physics_update_one_entity ;; Load routine's memory position
    ld b, #ent_cmp_movable                ;; Load filter mask (only physics entities)
    call man_for_all_matching
ret
