;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "render.h.s"
.include "cpctelera.h.s"
.include "../manager/entities.h.s"
.include "../manager/huditems.h.s"
.include "interruptions.h.s"
.include "cpct_override.h.s"

;; Sprites
.globl cpct_drawSprite_asm
.globl cpcto_drawSprite_20x25
.globl cpcto_drawSolidBox_20x25
;; Video
.globl cpct_getScreenPtr_asm
.globl cpcto_getScreenPtr_20x25
.globl cpct_setPALColour_asm
.globl cpct_setPalette_asm
.globl cpct_setVideoMode_asm

;; Import palette
.globl _g_palette

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Set Palette                                                      ;;
;;  Sets the game palette                                           ;;
;; Input: -                                                         ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_set_palette:
    ld hl, #_g_palette  ;; HL = Palette array
    ld de, #16          ;; Colours to copy (all of them)
    call cpct_setPalette_asm
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialize Render  For Game                                      ;;
;;  Initializes render configuration such as video mode or palette  ;;
;; Input: -                                                         ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_game_init:
    ;; Set video mode to 0 (1 byte = 2 pixels)
    ld c, #0x00
    call cpct_setVideoMode_asm

    ;; Set border
    ld l, #16
    ld h, #0x00 ;; Gray border
    call cpct_setPALColour_asm   ;;Input: HL, destroys AF, B requires firmware disabled
    
    ;; Set background to black
    ld l, #0x00
    ld h, #0x14 ;; Black background
    call cpct_setPALColour_asm   ;;Input: HL, destroys AF, B requires firmware disabled

    ;; Change screen size to 20x25 characters
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        ;; R1 = Screen width (in characters) = 0x14 (20)
        ld de, #0x0114
        call sys_render_change_CRTC
        ;cpctm_setCRTCReg 0x01, 0x14

        ;; R2 = Horizontal Sync Start        = 0x24 (36)
        ld de, #0x0224
        call sys_render_change_CRTC
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialize Render For Menu                                       ;;
;;  Initializes render configuration such as video mode or palette  ;;
;; Input: -                                                         ;;
;; Output: Destroys AF, BC and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_menu_init:

    ;; Set video mode to 0 (1 byte = 2 pixels)
    ld c, #0x00
    call cpct_setVideoMode_asm

    ;; Set border
    ld l, #16
    ld h, #0x14 ;; Black border
    call cpct_setPALColour_asm   ;;Input: HL, destroys F, BC and HL, requires firmware disabled
    
    ;; Assign Game Palette
    call sys_render_set_palette

    ;; Change screen size to 40x25 characters
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

        ;; R1 = Screen width (in characters) = 0x28 (40)
        ld de, #0x0128
        call sys_render_change_CRTC

        ;; R2 = Horizontal Sync Start        = 0x2E (46)
        ld de, #0x022E
        call sys_render_change_CRTC
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render One Entity                                                ;;
;;  Renders the specified entity, given its sprite and its position ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_one_entity:
    ;; Check erase by offset box or previous render position
    call sys_render_check_special_erasing
    jr nz, erase_by_prev_render

        ;; Erases by offset
        call sys_render_erase_by_offset
        jp erased_by_offset_box

    erase_by_prev_render:

        ;; Get previous drawing position (video memory) to erase the trail
        ld l, ent_prev_prev_render(ix)
        ld h, 1+ent_prev_prev_render(ix)

        ;; Erases the trail
        ex de, hl           ;; DE stores the video memory pointer
        ld c, ent_spr_w(ix) ;; C stores the Width
        ld b, ent_spr_h(ix) ;; B stores the height
        ld a, #0x00         ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25

    erased_by_offset_box:

    ;; Check if entity is marked for death to avoid rendering
    ld a, ent_type(ix)      ;; Load type in A
    and #ent_type_dead      ;; AND with death mask
    jr nz, render_one_skip  ;; If Not Zero, then death bit is marked and should NOT be rendered

    ;; Load X and Y pos, and start of video memory
    ld c, ent_x_pos(ix)         ;; X pos
    ld b, ent_y_pos(ix)         ;; Y pos
    draw_buffer = . + 2         ;; below instr is 11 00 C0, .+2 points to C0,
    ld de, #CPCT_VMEM_START_ASM ;; Video Memory start
    call cpcto_getScreenPtr_20x25;; Load screen pointer into HL

    ;; Save render position for next cycle's erase
        ;;load in DE previous 1 and then store it in previous 2, ez
    ld e, ent_prev_render(ix)           
    ld d, 1+ent_prev_render(ix)

    ld ent_prev_prev_render(ix), e
    ld 1+ent_prev_prev_render(ix), d 

    ld ent_prev_render(ix), l
    ld 1+ent_prev_render(ix), h

    ex de, hl                   ;; DE stores Video memory position (hl -> de)
    ld c, ent_spr_w(ix)         ;; C stores the width of the sprite
    ld b, ent_spr_h(ix)         ;; B stores the height of the sprite
    ld l, ent_sprite(ix)        ;; HL stores starting memory position of the sprite
    ld h, 1+ent_sprite(ix)
    call cpcto_drawSprite_20x25 ;; Draw sprite
    jp render_one_end

    ;; If it was dead, erase other buffer's trail
    render_one_skip:
        ld l, ent_prev_render(ix)
        ld h, 1+ent_prev_render(ix)

        ;; Erases the trail
        ex de, hl           ;; DE stores the video memory pointer
        ld c, ent_spr_w(ix) ;; C stores the Width
        ld b, ent_spr_h(ix) ;; B stores the height
        ld a, #0x00         ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25

        ld l, ent_prev_prev_render(ix)
        ld h, 1+ent_prev_prev_render(ix)

        ;; Erases the trail
        ex de, hl           ;; DE stores the video memory pointer
        ld c, ent_spr_w(ix) ;; C stores the Width
        ld b, ent_spr_h(ix) ;; B stores the height
        ld a, #0x00         ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25
    render_one_end:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render One Hud Item                                              ;;
;;  Renders the specified hud item, given its data                  ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_one_hud_item:
    ;; Check if hud item is marked for destruction to avoid rendering
    ld a, hud_type(ix)      ;; Load type in A
    and #hud_type_dead      ;; AND with death mask
    jr nz, render_hud_skip  ;; If Not Zero, then death bit is marked and should NOT be rendered

    ;; Check if it should be rendered
    ld a, hud_comps(ix)
    and #0x01           ;; Check render
    ret z               ;; If zero, don't render

    ld a, hud_comps(ix)
    and #hud_comp_alwaysdraw;; Get the "Always render" bit
    ld b, a             ;; B = Always Render bit
    
    ld a, hud_comps(ix)
    and #0b00001111     ;; Get the render components
    srl a               ;; Move the bits
    or b                ;; Recover the "Always render" bit
    ld b, a
    ld a, hud_comps(ix)
    and #0b11110000     ;; Get the not render components
    or b
    ld hud_comps(ix), a ;; Save

    ;; Load X and Y pos, and start of video memory
    ld c, hud_x_pos(ix)         ;; X pos
    ld b, hud_y_pos(ix)         ;; Y pos
    draw_buffer_hud = . + 2     ;; below instr is 11 00 C0, .+2 points to C0,
    ld de, #CPCT_VMEM_START_ASM ;; Video Memory start
    call cpcto_getScreenPtr_20x25;; Load screen pointer into HL

    ;; Save render position for next cycle's erase
        ;;load in DE previous 1 and then store it in previous 2, ez
    ld e, hud_prev_render(ix)           
    ld d, 1+hud_prev_render(ix)

    ld hud_prev_prev_render(ix), e
    ld 1+hud_prev_prev_render(ix), d 

    ld hud_prev_render(ix), l
    ld 1+hud_prev_render(ix), h

    ex de, hl                   ;; DE stores Video memory position (hl -> de)
    ld c, hud_spr_w(ix)         ;; C stores the width of the sprite
    ld b, hud_spr_h(ix)         ;; B stores the height of the sprite
    ld l, hud_sprite(ix)        ;; HL stores starting memory position of the sprite
    ld h, 1+hud_sprite(ix)
    call cpcto_drawSprite_20x25 ;; Draw sprite
    jp render_hud_end

    ;; If it was dead, erase other buffer's trail
    render_hud_skip:
        ;; Load prev pointer
        ld l, hud_prev_render(ix)
        ld h, 1+hud_prev_render(ix)

        ;; Erases the trail
        ex de, hl           ;; DE stores the video memory pointer
        ld c, hud_spr_w(ix) ;; C stores the Width
        ld b, hud_spr_h(ix) ;; B stores the height
        ld a, #0x00         ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25

        ;; Load prev prev pointer
        ld l, hud_prev_prev_render(ix)
        ld h, 1+hud_prev_prev_render(ix)

        ;; Erases the trail
        ex de, hl           ;; DE stores the video memory pointer
        ld c, hud_spr_w(ix) ;; C stores the Width
        ld b, hud_spr_h(ix) ;; B stores the height
        ld a, #0x00         ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25
    render_hud_end:
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render Menu Hud Item                                             ;;
;;  Renders the specified hud item, adapted (40x25 single buffer)   ;;
;; Input: IX with entity's memory position                          ;;
;; Output: Destroys AF, BC, DE and HL                               ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_menu_hud_item:
    ;; Load X and Y pos, and start of video memory
    ld c, hud_x_pos(ix)         ;; X pos
    ld b, hud_y_pos(ix)         ;; Y pos
    ld de, #CPCT_VMEM_START_ASM ;; Video Memory start
    call cpct_getScreenPtr_asm  ;; Load screen pointer into HL

    ex de, hl                   ;; DE stores Video memory position (hl -> de)
    ld c, hud_spr_w(ix)         ;; C stores the width of the sprite
    ld b, hud_spr_h(ix)         ;; B stores the height of the sprite
    ld l, hud_sprite(ix)        ;; HL stores starting memory position of the sprite
    ld h, 1+hud_sprite(ix)
    call cpct_drawSprite_asm    ;; Draw sprite
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render All Renderable Entities                                   ;;
;;  Asks the manager to render every renderable entity              ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_all_entites:
    ld hl, #sys_render_one_entity   ;; Load routine's memory position
    ld b, #ent_cmp_renderable       ;; Load filter mask (only renderable entities)
    call man_for_all_matching
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render All Hud Items                                             ;;
;;  Asks the manager to render every hud item                       ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_all_hud_items:
    ld hl, #sys_render_one_hud_item  ;; Load routine's memory position
    call man_hud_for_all
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render All Items                                                 ;;
;;  Asks the manager to render every hud item and renderable entity ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_all:
    call sys_render_all_entites
    call sys_render_all_hud_items

    ;;Double buffer magic: change the buffer
    call sys_render_SWAP_screen
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Render All Menu Hud Items                                        ;;
;;  Asks the manager to render every hud item for menus             ;;
;; Input: -                                                         ;;
;; Output: Destroys IX, A, B, DE and HL                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_all_menu_hud:
    ld hl, #sys_render_menu_hud_item  ;; Load routine's memory position
    call man_hud_for_all
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Change an CRTC Register                                          ;;
;;  Changes the specified CRTC register to the specified value      ;;
;; Input: D with register number, E with new register value         ;;
;; Output: Destroys BC                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_change_CRTC:
    ;;TODO : there is a macro for this
    ld  bc, #0xBC00
    out (c), d
    inc b   ;; B = 0xBD
    out (c), e
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Swap "screens"                                                   ;;
;;  Changes between "screens" (buffers actually)                    ;;
;; Input: -                                                         ;;
;; Output: Destroys A, DE and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_SWAP_screen:
    screen_swap_routine = . +1      ;;magic translocator
    jp sys_render_SWAP_to_C4
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Swap the render to C4                                            ;;
;;  change draw_buffer and screen_swap_routine                      ;;
;; Input: -                                                         ;;
;; Output: Destroys A, DE and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_SWAP_to_C4:
    ;;change register crtc to show (draw in) C400
    ld de, #0x0C32                  ;;change register 12 (VMEM) init to (C000) + (2 * 200) 
    call sys_render_change_CRTC     ;;                0C                   3          2 

    ;;tell program to draw in C000
    ld a, #0xC0                     ;; load C0 in a
    ld (draw_buffer), a             ;; set draw_buffer to C0
    ld (draw_buffer_hud), a         ;; set draw_buffer to C0

    ;;tell program to SWAP to the other buffer
    ld hl, #sys_render_SWAP_to_C0   ;; load in hl where is SWAP_to_C0
    ld (screen_swap_routine), hl    ;; reset translocator coordinates
    
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Swap the render to C0                                            ;;
;;  change draw_buffer and screen_swap_routine                      ;;
;; Input: -                                                         ;;
;; Output: Destroys A, DE and HL                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_SWAP_to_C0:
    ;;change register crtc to show (draw in) C000
    ld de, #0x0C30                  ;;change register 12 (VMEM) init to (C000) + (0 * 200) 
    call sys_render_change_CRTC     ;;                0C                   3        0 

    ;;tell program to draw in C000
    ld a, #0xC4                     ;; load C4 in a
    ld (draw_buffer), a             ;; set draw_buffer to C4
    ld (draw_buffer_hud), a         ;; set draw_buffer to C4

    ;;tell program to SWAP to the other buffer
    ld hl, #sys_render_SWAP_to_C4   ;; load in hl where is SWAP_to_C4
    ld (screen_swap_routine), hl    ;; reset translocator coordinates
    
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Wait a frame                                                      ;;
;;  Waits a frame using VSync and halts                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
wait_a_frame:
    ;; Wait until interruption counter gets to the desired amount
    ;; Interruption counter increases 300 times per second regardless of code executed
        ld a, (interrupt_counter)
        cp #interruptions_per_cycle+1
    jr c, wait_a_frame
    
    ;; Reset interruption counter
    call sys_interruption_reset_counter
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mark for Special Erase                                           ;;
;;  Mark an entity to use prev_prev_render to erase                 ;;
;; Input: IX with entity's memory                                   ;;
;; Output:  Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_mark_special_erasing:
    ;; Mark for special erasing
    ld a, ent_comps(ix)
    or #ent_cmp_erase_sp2
    or #ent_cmp_erase_sp
    ld ent_comps(ix), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rotate Special Erase                                             ;;
;;  Rotate special erasing bits and check for this frame            ;;
;; Input: IX with entity's memory                                   ;;
;; Output:  Z if normal erasing, NZ if special. Destroys A, B and D ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_check_special_erasing:
    ld d, #0 ;; D = 0

    ;; Get components and save in B
    ld a, ent_comps(ix)
    ld b, a

    ;; Check if special erase is needed this render
    and #ent_cmp_erase_sp
    jr z, skip_special_erase

        ;; Save result in D
        ld d, a ;; D != 0

        ;; Delete that bit flag (saved in B)
        ld a, b
        and #0xFF-ent_cmp_erase_sp
        ld b, a
    skip_special_erase:

    ;; Check if we need to rotate the second special erase bit
    ld a, b
    and #ent_cmp_erase_sp2
    jr z, skip_special_erase_rotation

        ;; Delete the second sp bit and set to 1 the other
        ld a, b
        and #0xFF-ent_cmp_erase_sp2
        or #ent_cmp_erase_sp
        ld b, a
    skip_special_erase_rotation:

    ;; Save B in entity's components
    ld ent_comps(ix), b

    ;; Return flag
    ld a, d ;; D has 0 if normal, !=0 if special
    or d    ;; Z if normal, NZ if special
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Erase By Offset                                                  ;;
;;  Make an (X,Y,W,H) box and erase with it                         ;;
;; Input: -                                                         ;;
;; Output:  Destroys ???                                            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_erase_by_offset:
    ;; Check erasing box
    call sys_render_get_erasing_box

    ;; Check if we should erase (width=E=0)
    ld a, e
    or a
    ret z
        ;; Get Screen pointer
        push de     ;; Save W and H in the stack
            ;; BC = X and Y pos
            ;; DE = Video memory start
        ld hl, #draw_buffer
        ld d, (hl)
        dec hl
        ld e, (hl)
        call cpcto_getScreenPtr_20x25;; Load screen pointer into HL

        ;; Draw the desired box
        ex de, hl   ;; Get the memory position in DE
        pop bc      ;; Pop the Width and Height from the stack
        ld a, #0    ;; A stores the color pattern
        call cpcto_drawSolidBox_20x25
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Get Erasing Box                                                  ;;
;;  Make an (X,Y,W,H) box using prevX and prevY to be optimal       ;;
;; Input: -                                                         ;;
;; Output:  C = X position                                          ;;
;;          B = Y position                                          ;;
;;          E = box width                                           ;;
;;          D = box height                                          ;;
;;          Destroys A                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_render_get_erasing_box:
    ;; Check if prev X = 0
    ld a, ent_prev_prev_x(ix)
    or a
    jr nz, check_prev_y_is_0

        ;; Check if prev Y = 0
        ld a, ent_prev_prev_y(ix)
        or a
            ;; Draw no box
            jp z, make_no_box
        ;; Check if prev Y < 0
        and #0x80
            ;; Box (x, y+prevY, w, !prevY)
            jp nz, make_box_x0_Yneg
        ;; Check if prev Y > 0
            ;; Box (x, y+h, w, prevY)
            jp make_box_x0_Ypos

    ;; Check if prev Y = 0
    check_prev_y_is_0:
    ld a, ent_prev_prev_y(ix)
    or a
    jp nz, sys_render_make_prev_prev_sprite_box ;; If prevX!=0 and prevY!=0, make whole box
        ;; Check if prev X < 0
        ld a, ent_prev_prev_x(ix)
        and #0x80
            ;; Box (x+prevX, y, !prevX, h)
            jp nz, make_box_y0_Xneg
        ;; Check if prev X > 0
            ;; Box (x+w, y, prevX, h)
            jp make_box_y0_Xpos

ret

sys_render_make_sprite_box:
    ld c, ent_x_pos(ix)
    ld b, ent_y_pos(ix)
    ld e, ent_spr_w(ix)
    ld d, ent_spr_h(ix)
ret

make_box_x0_Yneg:
    call sys_render_make_sprite_box
    ld a, ent_prev_prev_y(ix)   ;; | Add PrevY to Y
    add b                       ;; |
    ld b, a                     ;; |

    ld a, ent_prev_prev_y(ix)   ;; | Height = abs(prevY)
    neg                         ;; |
    ld d, a                     ;; | 
ret
make_box_x0_Ypos:
    call sys_render_make_sprite_box
    ld a, ent_spr_h(ix)         ;; | Add Height to Y
    add b                       ;; |
    ld b, a                     ;; |

    ld d, ent_prev_prev_y(ix)   ;; | Height = prevY
ret
make_box_y0_Xneg:
    call sys_render_make_sprite_box
    ld a, ent_prev_prev_x(ix)   ;; | Add PrevX to X
    add c                       ;; |
    ld c, a                     ;; |

    ld a, ent_prev_prev_x(ix)   ;; | Width = abs(prevX)
    neg                         ;; |
    ld e, a                     ;; | 
ret
make_box_y0_Xpos:
    call sys_render_make_sprite_box
    ld a, ent_spr_w(ix)         ;; | Add Width to X
    add c                       ;; |
    ld c, a                     ;; |

    ld e, ent_prev_prev_x(ix)   ;; | Width = prevX
ret
make_no_box:
    ld e, #0    ;; Width = 0 means don't draw box
ret

sys_render_make_prev_sprite_box:
    ld a, ent_x_pos(ix) ;; A = X pos
    ld b, ent_prev_x(ix);; B = prevX
    add b               ;; A = A - B = Xpos - prevX
    ld c, a             ;; C = Xpos - prevX

    ld a, ent_y_pos(ix) ;; A = Y pos
    ld b, ent_prev_y(ix);; B = prevY
    add b               ;; A = A - B = Ypos - prevY
    ld b, a             ;; B = Ypos - prevY

    ld e, ent_spr_w(ix) ;; E = Width
    ld d, ent_spr_h(ix) ;; D = Height
ret
sys_render_make_prev_prev_sprite_box:
    ld a, ent_x_pos(ix) ;; A = X pos
    ld b, ent_prev_prev_x(ix);; B = prevX
    add b               ;; A = A - B = Xpos - prevX
    ld c, a             ;; C = Xpos - prevX

    ld a, ent_y_pos(ix) ;; A = Y pos
    ld b, ent_prev_prev_y(ix);; B = prevY
    add b               ;; A = A - B = Ypos - prevY
    ld b, a             ;; B = Ypos - prevY

    ld e, ent_spr_w(ix) ;; E = Width
    ld d, ent_spr_h(ix) ;; D = Height
ret

sys_render_get_drawbuffer:
    ld hl, #draw_buffer
    ld d, (hl)
    dec hl
    ld e, (hl)
ret

sys_render_get_other_drawbuffer:
    ld hl, #draw_buffer
    ld d, (hl)
    dec hl
    ld e, (hl)

    ld a, d
    cp #0xC0
    jp nz, set_the_C0
        ld d, #0xC4
        ret
    set_the_C0:
        ld d, #0xC0
ret 
