;======================= LICENSE NOTICE =================================
;This file is part of HyperSpace, a videogame for Amstrad CPC 464.
;Copyright (C) 2020 HardBytes
;
;    HyperSpace is free software: you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.
;
;    HyperSpace is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with HyperSpace.  If not, see <https://www.gnu.org/licenses/>.
;========================================================================    

.include "score.h.s"
.include "cpctelera.h.s"
.include "../manager/huditems.h.s"
.include "../templates/templates.h.s"
.include "../sprites/sprites.h.s"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; VARIABLES                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
score_storage:: .ds score_digits
score_storage_end = . - 1

score_backup::  .ds score_digits
score_hiscore:: .ds score_digits

score_modified: .db 1
reached_top:    .db 0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Init points                                                      ;;
;;  Initializes points to zero                                      ;;
;; Input: -                                                         ;;
;; Output: Destroys BC, DE and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_init:
    ;; Set score to all zeros
    ld hl, #score_storage   ;; HL = Origin
    ld de, #score_storage+1 ;; DE = Destination
    ld bc, #score_digits-1  ;; BC = Number of bytes to copy
    ld (hl), #0             ;; Set first byte of origin to zero
    ldir                    ;; Copy that zero everywhere

    ;; Set backup to all zeros
    ld hl, #score_backup    ;; HL = Origin
    ld de, #score_backup+1  ;; DE = Destination
    ld bc, #score_digits-1  ;; BC = Number of bytes to copy
    ld (hl), #0             ;; Set first byte of origin to zero
    ldir                    ;; Copy that zero everywhere

    ;; Update modification
    ld a, #1
    ld (score_modified), a

    ;; Reset top
    ld a, #0
    ld (reached_top), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Add points                                                       ;;
;;  Adds the specified amount of score to the score counter         ;;
;; Input: DEBC as an 8 digit decimal number (but in hexadecimal)    ;;
;; Output: Destroys A and HL                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_add:
    ;; DEBC = DDEEBBCC = 8 digit number to add to score

    ;; Update modification
    ld a, #1
    ld (score_modified), a

    ;; HL points to storage end
    ld hl, #score_storage_end

    ;; Add 2 last digits of input (C) to score 2 last digits
    ld a, (hl)
    add c
    daa
    ld (hl), a

    ;; HL points to previous 2 digits
    dec hl

    ;; Add 2 digits of input (B) to score, with carry from previous
    ld a, (hl)
    adc b
    daa
    ld (hl), a

    ;; HL points to previous 2 digits
    dec hl

    ;; Add 2 digits of input (E) to score, with carry from previous
    ld a, (hl)
    adc e
    daa
    ld (hl), a

    ;; If carry, max amount was exceeded, so cap that to maximum
    call c, set_max_points

    ;; Check hiScore
    ld a, (reached_top)
    or a
    ret nz

    call check_hi_score_reached 
ret
set_max_points:
    ld hl, #score_storage
    ld de, #score_storage+1
    ld bc, #score_digits-1

    ld (hl), #0x99
    ldir
ret
check_hi_score_reached:
    call sys_score_check_hiscore_reached
    jr z, turn_off_hi_mark
        ;; Mark hiScore reached
        ld a, #1
        ld (reached_top), a

        call sys_score_create_top_mark
        ret
    turn_off_hi_mark:
        ld a, #0
        ld (reached_top), a
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Draw Score String                                                ;;
;;  Draws the score string on the screen (different positions)      ;;
;; Input: -                                                         ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_create:
    ld d, #score_string_x_start
    ld e, #0
    call sys_score_create_string

    ld hl, #score_storage
    call sys_score_update_string
ret
sys_score_draw:
    ld hl, #score_storage
    call sys_score_update_string
ret
sys_score_hiscore_mainmenu_draw:
    ld d, #mainmenu_score_string_x_start
    ld e, #mainmenu_hiscore_string_y
    call sys_score_create_string

    ld hl, #score_hiscore
    call sys_score_update_string
ret
sys_score_hiscore_wingame_draw:
    ld d, #end_game_score_string_x_start
    ld e, #win_game_hiscore_string_y
    call sys_score_create_string

    ld hl, #score_hiscore
    call sys_score_update_string
ret
sys_score_end_game_draw:
    ld d, #end_game_score_string_x_start
    ld e, #end_game_score_string_y
    call sys_score_create_string

    ld hl, #score_storage
    call sys_score_update_string
ret
sys_score_win_game_draw:
    ld d, #end_game_score_string_x_start
    ld e, #win_game_score_string_y
    call sys_score_create_string

    ld hl, #score_storage
    call sys_score_update_string
ret

sys_score_create_top_mark:
    ld a, (reached_top)
    or a
    ret z
        push ix
        ld hl, #hud_template_topscore
        call man_create_one_hud_item
        pop ix
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create Score String                                              ;;
;;  Creates the score string in the HUD                             ;;
;; Input: D with starting X point and E with Y position             ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_create_string:
    ;; Update modification
    ld a, #1
    ld (score_modified), a

    ;; Save X and Y position
    push de

    ;; Delete previous score string
    ld b, #hud_type_score
    call man_mark_for_destroy_hud_of_type

    ;; Init counter
    ld a, #score_digits*2

    ;; Recover X and Y positions
    pop de
    ;; D = starting X point
    ;; E = Y position
    draw_loop:
        ;; Save counter
        push af

        push de
        ld hl, #hud_template_score_digit    ;; Template of extra life counter
        call man_create_one_hud_item
        pop de

        ;; Positionate
        ld hud_y_pos(ix), e
        ld hud_x_pos(ix), d
        ld a, #3
        add d
        ld d, a

        ;; Recover counter
        pop af

        dec a   ;; Decrease counter
    jp nz, draw_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Update Score String                                              ;;
;;  Updates the score string in the HUD with the current score      ;;
;; Input: -                                                         ;;
;; Output: Destroys ???                                             ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_update_string:
    ;; Check if string was modified since last time
    ld a, (score_modified)
    or a
    ret z
        ;; Turn off modification
        ld a, #0
        ld (score_modified), a

        ;; Search score string
        ld b, #hud_type_score
        call man_hud_search_type
        ;; IX = first two digits

        ;; Set HL to score beggining
        ;; HL = score_storage

        ;; Init counter
        ld a, #score_digits

        draw_update_loop:
            ;; Save counter and score
            push af
            push hl

            ;; Get first digit
            ld a, (hl)
            rra
            rra
            rra
            rra
            and #0x0F
            call sys_score_get_number_sprite

            ;; Set redraw
            ld a, hud_comps(ix)
            or #hud_comp_redraw1
            or #hud_comp_redraw2
            or #hud_comp_redraw3
            ld hud_comps(ix), a

            ;; Go to next digit
            ld bc, #hud_item_size
            add ix, bc

            ;; Recover score
            pop hl
            push hl

            ;; Get second digit
            ld a, (hl)                          ;; Value to show
            and #0x0F
            call sys_score_get_number_sprite

            ;; Set redraw
            ld a, hud_comps(ix)
            or #hud_comp_redraw1
            or #hud_comp_redraw2
            or #hud_comp_redraw3
            ld hud_comps(ix), a

            ;; Go to next digit
            ld bc, #hud_item_size
            add ix, bc

            ;; Recover counter and score
            pop hl
            pop af

            inc hl  ;; Point to next two digits
            dec a   ;; Decrease counter
        jp nz, draw_update_loop
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Get Number Sprite                                                ;;
;;  Gets in IX a HUD Element with the number specified (1 digit 0-9);;
;; Input: A with value (0-9), IX with target HUD item               ;;
;; Output: IX with HUD item memory                                  ;;
;;         Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_get_number_sprite:
    ;; Select number sprite according to remaining extra lifes
    ld hl, #_spr_lifecount_00       ;; Load first sprite (0 lifes)
    ld bc, #spr_lifecount_w*spr_lifecount_h

    extralife_count_hud_loop:
    dec a                           ;; Decrease life counter
    jp m, finish_extralife_hud_creation ;; If no more extralifes remain, assign chosen sprite
        add hl, bc                  ;; Advance to next sprite
        jp extralife_count_hud_loop ;; Repeat life check

    ;; Sprite chosen, now assign number sprite
    finish_extralife_hud_creation:
    ld hud_sprite(ix)  , l  ;; Assign new sprite
    ld 1+hud_sprite(ix), h
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Save Score Backup                                                ;;
;;  Save current score on backup bytes                              ;;
;; Input: -                                                         ;;
;; Output: Destroys BC, DE and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_save_score_state:
    ;; Copy HL to DE: Save current score on backup
    ld hl, #score_storage
    ld de, #score_backup
    ld bc, #score_digits
    ldir
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Load Score Backup                                                ;;
;;  Load backup on current score                                    ;;
;; Input: -                                                         ;;
;; Output: Destroys BC, DE and HL                                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_load_score_state:
    ;; Copy HL to DE: Load current backup to score
    ld hl, #score_backup
    ld de, #score_storage
    ld bc, #score_digits
    ldir
ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Save HiScore                                                     ;;
;;  Save current score on hiscore if it's higher                    ;;
;; Input: -                                                         ;;
;; Output: Destroys A, BC, DE and HL                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sys_score_check_save_hiscore:
    call sys_score_check_hiscore_reached
    ret z

        ;; Set hiscore
        ;; Copy HL to DE: Save current score on hiScore
        ld hl, #score_storage
        ld de, #score_hiscore
        ld bc, #score_digits
        ldir
ret

sys_score_check_hiscore_reached:
    ;; Check if Score is higher than HiScore
    ld hl, #score_storage
    ld de, #score_hiscore

    ld a, #score_digits

    repeat_hiscore_check:
    ld__iyl_a
    ld a, (hl)
    ex de, hl
    ld b, (hl)
    cp b        ;; A - B
    ld__a_iyl
    jp c, hiscore_not_reached   ;; If carry, hiScore is higher than score, finish

        ;; If it's higher, jump to set hiScore
        jr nz, hiscore_reached

        ;; If this is the last byte, skip check
        dec a
        jp z, hiscore_not_reached

        ;; Else try again with next byte
        inc hl
        ex de, hl
        inc hl
        jr repeat_hiscore_check

    hiscore_reached:
        ld a, #1
        or a
        ;; NZ
        ret

    hiscore_not_reached: 
        ld a, #0
        or a
        ;; Z
ret
